import os
import asyncio
import pathlib
import multiprocessing
import synapse.exc as s_exc
import synapse.glob as s_glob
import synapse.common as s_common

from unittest.mock import patch

import synapse.lib.base as s_base
import synapse.lib.const as s_const
import synapse.lib.lmdbslab as s_lmdbslab
import synapse.lib.thisplat as s_thisplat

import synapse.tests.utils as s_t_utils
from synapse.tests.utils import alist

def getFileMapCount(filename):
    filename = str(filename)
    count = 0
    with open(f'/proc/{os.getpid()}/maps') as maps:
        for line in maps:
            if len(line) < 50:
                continue
            if line.rstrip().endswith(filename):
                count += 1
    return count

class LmdbSlabTest(s_t_utils.SynTest):
    def __init__(self, *args, **kwargs):
        self._nowtime = 1000
        s_t_utils.SynTest.__init__(self, *args, **kwargs)

    async def test_lmdbslab_scankeys(self):

        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')
            async with await s_lmdbslab.Slab.anit(path) as slab:

                testdb = slab.initdb('test')
                dupsdb = slab.initdb('dups', dupsort=True)
                editdb = slab.initdb('edit')

                self.eq((), list(slab.scanKeys(db=testdb)))
                self.eq((), list(slab.scanByDupsBack(b'asdf', db=dupsdb)))

                slab.put(b'hehe', b'haha', db=dupsdb)
                slab.put(b'hehe', b'lolz', db=dupsdb)
                slab.put(b'hoho', b'asdf', db=dupsdb)

                self.eq((), list(slab.scanByDupsBack(b'h\x00', db=dupsdb)))

                slab.put(b'hehe', b'haha', db=testdb)
                slab.put(b'hoho', b'haha', db=testdb)

                testgenr = slab.scanKeys(db=testdb)
                dupsgenr = slab.scanKeys(db=dupsdb)

                testlist = [next(testgenr)]
                dupslist = [next(dupsgenr)]

                slab.put(b'derp', b'derp', db=editdb)

                # bump them both...
                await s_lmdbslab.Slab.syncLoopOnce()

                testlist.extend(testgenr)
                dupslist.extend(dupsgenr)

                self.eq(testlist, (b'hehe', b'hoho'))
                self.eq(dupslist, (b'hehe', b'hehe', b'hoho'))

                # now lets delete the key we're on
                testgenr = slab.scanKeys(db=testdb)
                dupsgenr = slab.scanKeys(db=testdb)

                testlist = [next(testgenr)]
                dupslist = [next(dupsgenr)]

                slab.delete(b'hehe', db=testdb)
                for lkey, lval in slab.scanByDups(b'hehe', db=dupsdb):
                    slab.delete(lkey, lval, db=dupsdb)

                await s_lmdbslab.Slab.syncLoopOnce()

                testlist.extend(testgenr)
                dupslist.extend(dupsgenr)

                self.eq(testlist, (b'hehe', b'hoho'))
                self.eq(dupslist, (b'hehe', b'hoho'))

                # by pref
                self.eq([b'hoho'], list(slab.scanKeysByPref(b'h', db=dupsdb)))
                self.eq([], list(slab.scanKeysByPref(b'z', db=dupsdb)))

    async def test_lmdbslab_base(self):

        with self.getTestDir() as dirn0, self.getTestDir(startdir=dirn0) as dirn:

            path = os.path.join(dirn, 'test.lmdb')

            await self.asyncraises(s_exc.BadArg, s_lmdbslab.Slab.anit(path, map_size=None))

            slab = await s_lmdbslab.Slab.anit(path, map_size=1000000, lockmemory=True)

            slabs = slab.getSlabsInDir(dirn)
            self.eq(slabs, [slab])

            slabs = slab.getSlabsInDir(pathlib.Path(dirn) / 'nowhere')
            self.len(0, slabs)

            slabs = slab.getSlabsInDir(pathlib.Path(dirn).parent)
            self.ge(1, len(slabs))

            foo = slab.initdb('foo')
            baz = slab.initdb('baz')
            bar = slab.initdb('bar', dupsort=True)
            empty = slab.initdb('empty')
            barfixed = slab.initdb('barfixed', dupsort=True, dupfixed=True)

            slab.put(b'\x00\x01', b'hehe', db=foo)
            slab.put(b'\x00\x02', b'haha', db=foo)
            slab.put(b'\x01\x03', b'hoho', db=foo)

            for db in (bar, barfixed):
                slab.put(b'\x00\x01', b'hehe', dupdata=True, db=db)
                slab.put(b'\x00\x02', b'haha', dupdata=True, db=db)
                slab.put(b'\x00\x02', b'visi', dupdata=True, db=db)
                slab.put(b'\x00\x02', b'zomg', dupdata=True, db=db)
                slab.put(b'\x00\x03', b'hoho', dupdata=True, db=db)

            slab.put(b'\x00\x01', b'hehe', db=baz)
            slab.put(b'\xff', b'haha', db=baz)

            slab.put(b'\xff\xff', b'hoho', append=True, db=baz)  # Should succeed
            slab.put(b'\xaa\xff', b'hoho', append=True, db=baz)  # Should fail (not the last key)

            self.true(slab.dirty)

            self.true(slab.forcecommit())
            self.false(slab.dirty)

            self.eq(b'\x00\x01', slab.firstkey(db=foo))
            self.none(slab.firstkey(db=empty))
            self.eq(b'\x01\x03', slab.lastkey(db=foo))
            self.none(slab.lastkey(db=empty))

            self.eq(b'hehe', slab.get(b'\x00\x01', db=foo))

            items = list(slab.scanByPref(b'\x00', db=foo))
            self.eq(items, ((b'\x00\x01', b'hehe'), (b'\x00\x02', b'haha')))

            items = list(slab.scanByRange(b'\x00\x02', b'\x01\x03', db=foo))
            self.eq(items, ((b'\x00\x02', b'haha'), (b'\x01\x03', b'hoho')))

            for db in (bar, barfixed):
                items = list(slab.scanByDups(b'\x00\x02', db=db))
                self.eq(items, ((b'\x00\x02', b'haha'), (b'\x00\x02', b'visi'), (b'\x00\x02', b'zomg')))

                items = list(slab.scanByDups(b'\x00\x04', db=db))
                self.eq(items, ())

            # Test scanByPref startkey, startvalu
            items = list(slab.scanByPref(b'\x00', db=bar))
            alld = [(b'\x00\x01', b'hehe'),
                    (b'\x00\x02', b'haha'),
                    (b'\x00\x02', b'visi'),
                    (b'\x00\x02', b'zomg'),
                    (b'\x00\x03', b'hoho')]
            self.eq(alld, items)

            items = list(slab.scanByPref(b'\x00', startkey=b'\x02', db=bar))
            self.eq(alld[1:], items)

            items = list(slab.scanByPref(b'\x00', startkey=b'\x02', startvalu=b'vaaa', db=bar))
            self.eq(alld[2:], items)

            self.true(slab.prefexists(b'\x00', db=baz))
            self.true(slab.prefexists(b'\x00\x01', db=baz))
            self.false(slab.prefexists(b'\x00\x03', db=baz))
            self.false(slab.prefexists(b'\x02', db=baz))
            self.true(slab.prefexists(b'\xff\xff', db=baz))
            self.false(slab.prefexists(b'\xff\xff', db=foo))
            self.false(slab.prefexists(b'\xaa\xff', db=baz))

            self.true(slab.rangeexists(b'\x00', b'\x01', db=baz))
            self.true(slab.rangeexists(b'\x00\x00', b'\x00\x04', db=baz))
            self.false(slab.rangeexists(b'\x00\x04', b'\x01', db=baz))
            self.true(slab.rangeexists(b'\x05', None, db=baz))
            self.false(slab.rangeexists(b'\xfa', b'\xfc', db=baz))
            self.false(slab.rangeexists(b'\x00\x00', b'\x00\x00', db=foo))
            self.false(slab.rangeexists(b'\x01\x04', b'\x01\x05', db=foo))

            # backwards scan tests
            items = list(slab.scanByRangeBack(b'\x00', db=foo))
            self.eq(items, ())

            items = list(slab.scanByPrefBack(b'\x00', db=foo))
            self.eq(items, ((b'\x00\x02', b'haha'), (b'\x00\x01', b'hehe')))

            items = list(slab.scanByPrefBack(b'\x01', db=foo))
            self.eq(items, ((b'\x01\x03', b'hoho'),))

            items = list(slab.scanByPrefBack(b'\xff', db=baz))
            self.eq(items, ((b'\xff\xff', b'hoho'), (b'\xff', b'haha')))

            items = list(slab.scanByRangeBack(b'\x00\x03', db=foo))
            self.eq(items, ((b'\x00\x02', b'haha'), (b'\x00\x01', b'hehe')))

            items = list(slab.scanByRangeBack(b'\x00\x03', b'\x00\x02', db=foo))
            self.eq(items, ((b'\x00\x02', b'haha'), ))

            items = list(slab.scanByRangeBack(b'\x01\x03', b'\x00\x02', db=foo))
            self.eq(items, ((b'\x01\x03', b'hoho'), (b'\x00\x02', b'haha')))

            items = list(slab.scanByRangeBack(b'\x01\x05', b'\x00\x02', db=foo))
            self.eq(items, ((b'\x01\x03', b'hoho'), (b'\x00\x02', b'haha')))

            for db in (bar, barfixed):
                items = list(slab.scanByDupsBack(b'\x00\x02', db=db))
                self.eq(items, ((b'\x00\x02', b'zomg'), (b'\x00\x02', b'visi'), (b'\x00\x02', b'haha')))

                items = list(slab.scanByDupsBack(b'\x00\x04', db=db))
                self.eq(items, ())

                with s_lmdbslab.ScanBack(slab, db=db) as scan:
                    scan.first()
                    self.eq(scan.atitem, (b'\x00\x03', b'hoho'))

            items = list(slab.scanByFullBack(db=foo))
            self.eq(items, ((b'\x01\x03', b'hoho'), (b'\x00\x02', b'haha'), (b'\x00\x01', b'hehe')))

            with s_lmdbslab.ScanBack(slab, db=foo) as scan:
                scan.set_key(b'\x00\x02')
                self.eq(scan.atitem, (b'\x00\x02', b'haha'))

            # test scans on emptydb

            emptydb = slab.initdb('empty')

            items = list(slab.scanByPrefBack(b'\x00\x01', db=emptydb))
            self.eq(items, ())

            items = list(slab.scanByPrefBack(b'\xff\xff', db=emptydb))
            self.eq(items, ())

            items = list(slab.scanByRangeBack(b'\x00\x01', db=emptydb))
            self.eq(items, ())

            items = list(slab.scanByFullBack(db=emptydb))
            self.eq(items, ())

            # ok... lets start a scan and then rip out the xact...
            scan = slab.scanByPref(b'\x00', db=foo)
            self.eq((b'\x00\x01', b'hehe'), next(scan))

            slab.forcecommit()

            items = list(scan)
            self.eq(items, ((b'\x00\x02', b'haha'),))

            for db in (bar, barfixed):
                # to test iternext_dup, lets do the same with a dup scan
                scan = slab.scanByDups(b'\x00\x02', db=db)
                self.eq((b'\x00\x02', b'haha'), next(scan))

            slab.forcecommit()

            items = list(scan)
            self.eq(items, ((b'\x00\x02', b'visi'), (b'\x00\x02', b'zomg')))

            # do the same with backwards scanning
            scan = slab.scanByRangeBack(b'\x01\x03', db=foo)
            self.eq((b'\x01\x03', b'hoho'), next(scan))

            slab.forcecommit()

            items = list(scan)
            self.eq(items, ((b'\x00\x02', b'haha'), (b'\x00\x01', b'hehe')))

            # Copy a database inside the same slab
            self.raises(s_exc.DataAlreadyExists, slab.copydb, foo, slab, 'bar')
            self.eq(3, slab.copydb(foo, slab, 'foo2'))

            # Increase the size of the new source DB to trigger a resize on the next copydb
            foo2 = slab.initdb('foo2')
            slab.put(b'bigkey', b'x' * 1024 * 1024, dupdata=True, db=foo2)

            vardict = {}

            def progfunc(count):
                vardict['prog'] = count

            # Copy a database to a different slab
            path2 = os.path.join(dirn, 'test2.lmdb')
            async with await s_lmdbslab.Slab.anit(path2, map_size=512 * 1024) as slab2:
                with patch('synapse.lib.lmdbslab.PROGRESS_PERIOD', 2):

                    self.eq(4, slab.copydb(foo2, slab2, destdbname='foo2', progresscb=progfunc))
                    self.gt(vardict.get('prog', 0), 0)

            # Test slab.drop and slab.dbexists
            self.true(slab.dbexists('foo2'))
            slab.dropdb('foo2')
            self.false(slab.dbexists('foo2'))

            self.none(slab.dropdb('notadb'))

            # start a scan and then fini the whole db...
            scan = slab.scanByPref(b'\x00', db=foo)
            self.eq((b'\x00\x01', b'hehe'), next(scan))

            scanback = slab.scanByPrefBack(b'\x00', db=foo)
            self.eq((b'\x00\x02', b'haha'), next(scanback))

            await slab.fini()

            self.raises(s_exc.IsFini, next, scan)
            self.raises(s_exc.IsFini, next, scanback)

            slabs = s_lmdbslab.Slab.getSlabsInDir(dirn)
            self.len(0, slabs)

            # Ensure that our envar override for memory locking is acknowledged
            with self.setTstEnvars(SYN_LOCKMEM_DISABLE='1'):
                slab = await s_lmdbslab.Slab.anit(path, map_size=1000000, lockmemory=True)
                self.false(slab.lockmemory)
                self.none(slab.memlocktask)

    def simplenow(self):
        self._nowtime += 1000
        return self._nowtime

    async def test_lmdbslab_commit_warn(self):
        with self.getTestDir() as dirn, patch('synapse.lib.lmdbslab.Slab.WARN_COMMIT_TIME_MS', 1), \
                patch('synapse.common.now', self.simplenow):
            path = os.path.join(dirn, 'test.lmdb')
            with self.getAsyncLoggerStream('synapse.lib.lmdbslab', 'Commit with') as stream:
                async with await s_lmdbslab.Slab.anit(path, map_size=100000) as slab:
                    foo = slab.initdb('foo', dupsort=True)
                    byts = b'\x00' * 256
                    for i in range(10):
                        slab.put(b'\xff\xff\xff\xff' + s_common.guid(i).encode('utf8'), byts, db=foo)
                self.true(await stream.wait(timeout=1))

    async def test_lmdbslab_max_replay(self):
        with self.getTestDir() as dirn:
            path = os.path.join(dirn, 'test.lmdb')

            my_maxlen = 100

            # Make sure that we don't confuse the periodic commit with the max replay log commit

            with patch('synapse.lib.lmdbslab.Slab.COMMIT_PERIOD', 10):
                async with await s_lmdbslab.Slab.anit(path, map_size=100000, max_replay_log=my_maxlen) as slab:
                    foo = slab.initdb('foo', dupsort=True)
                    byts = b'\x00' * 256

                    waiter = s_base.Waiter(slab, 1, 'commit')

                    for i in range(150):
                        slab.put(b'\xff\xff\xff\xff' + s_common.guid(i).encode('utf8'), byts, db=foo)

                    self.true(slab.syncevnt.is_set())

                    retn = await waiter.wait(timeout=1)
                    self.nn(retn)
                    self.len(1, retn)

    async def test_lmdbslab_maxsize(self):
        with self.getTestDir() as dirn:
            path = os.path.join(dirn, 'test.lmdb')

            my_maxsize = 400000
            async with await s_lmdbslab.Slab.anit(path, map_size=100000, maxsize=my_maxsize) as slab:
                foo = slab.initdb('foo', dupsort=True)
                byts = b'\x00' * 256

                # Trigger an out-of-space
                with self.raises(s_exc.DbOutOfSpace):

                    for i in range(400):
                        slab.put(b'\xff\xff\xff\xff' + s_common.guid(i).encode('utf8'), byts, db=foo)

            # lets ensure our maxsize persisted and it caps the mapsize
            async with await s_lmdbslab.Slab.anit(path, map_size=100000, readonly=True) as newdb:
                self.eq(my_maxsize, newdb.mapsize)
                self.eq(my_maxsize, newdb.maxsize)

    async def test_lmdbslab_scanbump(self):

        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')

            async with await s_lmdbslab.Slab.anit(path, map_size=100000, growsize=10000) as slab:

                foo = slab.initdb('foo', dupsort=True)
                foo2 = slab.initdb('foo2', dupsort=False)
                bar = slab.initdb('bar', dupsort=True)

                multikey = b'\xff\xff\xff\xfe' + s_common.guid(2000).encode('utf8')

                byts = b'\x00' * 256
                for i in range(10):
                    slab.put(multikey, s_common.int64en(i), dupdata=True, db=foo)
                    slab.put(s_common.int64en(i), byts, db=foo2)

                iter1 = slab.scanByDups(multikey, db=foo)
                iter2 = slab.scanByFull(db=foo2)

                for _ in range(6):
                    next(iter1)
                    next(iter2)

                iterback = slab.scanByDupsBack(multikey, db=foo)
                next(iterback)

                iterback2 = slab.scanByFullBack(db=foo2)
                next(iterback2)

                iterback3 = slab.scanByDupsBack(multikey, db=foo)
                iterback4 = slab.scanByFullBack(db=foo2)

                for _ in range(8):
                    next(iterback3)
                    next(iterback4)

                iterback5 = slab.scanByDupsBack(multikey, db=foo)
                next(iterback5)

                iterback6 = slab.scanByFullBack(db=foo2)
                next(iterback6)

                # Delete keys to cause set_range in iternext to fail
                for i in range(5):
                    slab.delete(multikey, s_common.int64en(i + 5), db=foo)
                    slab.delete(s_common.int64en(i + 5), db=foo2)

                slab.forcecommit()

                self.raises(StopIteration, next, iter1)
                self.raises(StopIteration, next, iter2)
                self.len(5, iterback)
                self.len(5, iterback2)

                # Delete all the keys in front of a backwards scan
                for i in range(4):
                    slab.delete(multikey, s_common.int64en(i), db=foo)
                    slab.delete(s_common.int64en(i), db=foo2)

                self.raises(StopIteration, next, iterback3)
                self.raises(StopIteration, next, iterback4)

                # Delete remaining keys so curs.last fails
                slab.delete(multikey, s_common.int64en(4), db=foo)
                slab.delete(s_common.int64en(4), db=foo2)

                self.raises(StopIteration, next, iterback5)
                self.raises(StopIteration, next, iterback6)

                slab.put(b'\x00', b'asdf', dupdata=True, db=bar)
                slab.put(b'\x01', b'qwer', dupdata=True, db=bar)
                iterback = slab.scanByRangeBack(b'\x00', db=bar)
                self.eq((b'\x00', b'asdf'), next(iterback))
                slab.delete(b'\x00', b'asdf', db=bar)
                slab.forcecommit()
                self.raises(StopIteration, next, iterback)

                # range scan where we delete the entry we're on
                # and it's the only thing in the slab.
                iterrange = slab.scanByRange(b'\x00', db=bar)
                self.eq((b'\x01', b'qwer'), next(iterrange))
                slab.delete(b'\x01', b'qwer', db=bar)
                slab.forcecommit()
                self.raises(StopIteration, next, iterrange)

    async def test_lmdbslab_scanbump2(self):

        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')

            async with await s_lmdbslab.Slab.anit(path, map_size=100000, growsize=10000) as slab:

                dupydb = slab.initdb('dup', dupsort=True)
                dupndb = slab.initdb('ndup', dupsort=False)

                for db in (dupndb, dupydb):
                    slab.put(b'1', b'', db=db)
                    slab.put(b'2', b'', db=db)
                    slab.put(b'3', b'', db=db)

                    # forwards, bump after 2nd entry
                    it = slab.scanByFull(db=db)
                    self.eq((b'1', b''), next(it))
                    self.eq((b'2', b''), next(it))
                    slab.forcecommit()
                    self.eq((b'3', b''), next(it))
                    self.raises(StopIteration, next, it)

                    # backwards, bump after 2nd entry
                    it = slab.scanByFullBack(db=db)
                    self.eq((b'3', b''), next(it))
                    self.eq((b'2', b''), next(it))
                    slab.forcecommit()
                    self.eq((b'1', b''), next(it))
                    self.raises(StopIteration, next, it)

                    # forwards, bump/delete after 2nd entry
                    it = slab.scanByFull(db=db)
                    self.eq((b'1', b''), next(it))
                    self.eq((b'2', b''), next(it))
                    slab.forcecommit()
                    slab.delete(b'2', db=db)
                    self.eq((b'3', b''), next(it))
                    self.raises(StopIteration, next, it)

                    it = slab.scanByFull(db=db)
                    self.eq((b'1', b''), next(it))
                    slab.forcecommit()
                    slab.delete(b'3', db=db)
                    self.raises(StopIteration, next, it)

                    slab.put(b'2', b'', db=db)
                    slab.put(b'3', b'', db=db)

                    # backwards, bump/delete after 2nd entry
                    it = slab.scanByFullBack(db=db)
                    self.eq((b'3', b''), next(it))
                    self.eq((b'2', b''), next(it))
                    slab.forcecommit()
                    slab.delete(b'2', db=db)
                    self.eq((b'1', b''), next(it))
                    self.raises(StopIteration, next, it)

                    it = slab.scanByFullBack(db=db)
                    slab.forcecommit()
                    slab.delete(b'3', db=db)
                    self.eq((b'1', b''), next(it))
                    self.raises(StopIteration, next, it)

                slab.delete(b'1', db=dupydb)
                slab.delete(b'2', db=dupydb)
                slab.delete(b'3', db=dupydb)
                slab.put(b'0', b'', db=dupydb)
                slab.put(b'1', b'1', db=dupydb)
                slab.put(b'1', b'2', db=dupydb)
                slab.put(b'1', b'3', db=dupydb)
                slab.put(b'2', b'', db=dupydb)

                # dupsort=yes, forwards, same keys, bump after 2nd entry
                it = slab.scanByFull(db=dupydb)
                self.eq((b'0', b''), next(it))
                self.eq((b'1', b'1'), next(it))
                self.eq((b'1', b'2'), next(it))
                slab.forcecommit()
                self.eq((b'1', b'3'), next(it))
                self.eq((b'2', b''), next(it))
                self.raises(StopIteration, next, it)

                # forwards, bump/delete after 2nd entry
                it = slab.scanByFull(db=dupydb)
                self.eq((b'0', b''), next(it))
                self.eq((b'1', b'1'), next(it))
                slab.forcecommit()
                slab.delete(b'1', val=b'2', db=dupydb)
                self.eq((b'1', b'3'), next(it))
                self.eq((b'2', b''), next(it))
                self.raises(StopIteration, next, it)

                it = slab.scanByFull(db=dupydb)
                self.eq((b'0', b''), next(it))
                self.eq((b'1', b'1'), next(it))
                self.eq((b'1', b'3'), next(it))
                slab.forcecommit()
                slab.delete(b'1', val=b'3', db=dupydb)
                self.eq((b'2', b''), next(it))
                self.raises(StopIteration, next, it)

                slab.put(b'1', b'2', db=dupydb)
                slab.put(b'1', b'3', db=dupydb)

                # dupsort=yes, backwards, same keys, bump after 2nd entry
                it = slab.scanByFullBack(db=dupydb)
                self.eq((b'2', b''), next(it))
                self.eq((b'1', b'3'), next(it))
                self.eq((b'1', b'2'), next(it))
                slab.forcecommit()
                self.eq((b'1', b'1'), next(it))
                self.eq((b'0', b''), next(it))
                self.raises(StopIteration, next, it)

                # dupsort=yes, backwards, same keys, bump/delete after 2nd entry
                it = slab.scanByFullBack(db=dupydb)
                self.eq((b'2', b''), next(it))
                self.eq((b'1', b'3'), next(it))
                self.eq((b'1', b'2'), next(it))
                slab.forcecommit()
                slab.delete(b'1', val=b'2', db=dupndb)
                self.eq((b'1', b'1'), next(it))
                self.eq((b'0', b''), next(it))
                self.raises(StopIteration, next, it)

                slab.put(b'1', b'2', db=dupydb)
                slab.put(b'1', b'3', db=dupydb)

                # single key, forwards, bump after 2nd entry
                it = slab.scanByDups(db=dupydb, lkey=b'1')
                self.eq((b'1', b'1'), next(it))
                self.eq((b'1', b'2'), next(it))
                slab.forcecommit()
                self.eq((b'1', b'3'), next(it))
                self.raises(StopIteration, next, it)

                # single key, forwards, bump/delete after 2nd entry
                it = slab.scanByDups(db=dupydb, lkey=b'1')
                self.eq((b'1', b'1'), next(it))
                slab.forcecommit()
                slab.delete(b'1', val=b'2', db=dupydb)
                self.eq((b'1', b'3'), next(it))
                self.raises(StopIteration, next, it)

                it = slab.scanByDups(db=dupydb, lkey=b'1')
                self.eq((b'1', b'1'), next(it))
                slab.forcecommit()
                slab.delete(b'1', val=b'3', db=dupydb)
                self.raises(StopIteration, next, it)

                slab.put(b'1', b'2', db=dupydb)
                slab.put(b'1', b'3', db=dupydb)

                # dupsort=yes, backwards, same keys, bump after 2nd entry
                it = slab.scanByDupsBack(db=dupydb, lkey=b'1')
                self.eq((b'1', b'3'), next(it))
                self.eq((b'1', b'2'), next(it))
                slab.forcecommit()
                self.eq((b'1', b'1'), next(it))
                self.raises(StopIteration, next, it)

                # dupsort=yes, backwards, same keys, bump/delete after 2nd entry
                it = slab.scanByDupsBack(db=dupydb, lkey=b'1')
                self.eq((b'1', b'3'), next(it))
                self.eq((b'1', b'2'), next(it))
                slab.forcecommit()
                slab.delete(b'1', val=b'2', db=dupndb)
                self.eq((b'1', b'1'), next(it))
                self.raises(StopIteration, next, it)

    async def test_lmdbslab_count_empty(self):

        with self.getTestDir() as dirn:
            path = os.path.join(dirn, 'test.lmdb')
            async with await s_lmdbslab.Slab.anit(path, map_size=100000, growsize=10000) as slab:
                self.eq(0, await slab.countByPref(b'asdf'))

    async def test_lmdbslab_grow(self):

        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')

            async with await s_lmdbslab.Slab.anit(path, map_size=100000, growsize=10000) as slab:

                foo = slab.initdb('foo', dupsort=True)
                foo2 = slab.initdb('foo2', dupsort=False)

                byts = b'\x00' * 256
                for i in range(100):
                    slab.put(s_common.guid(i).encode('utf8'), byts, db=foo)
                    slab.put(s_common.guid(1000 + i).encode('utf8'), byts, db=foo2)

                count = 0
                for _, _ in slab.scanByRange(b'', db=foo):
                    count += 1
                self.eq(count, 100)

                count = 0
                for _, _ in slab.scanByRangeBack(b'ffffffffffffffffffffffffffffffff', db=foo):
                    count += 1
                self.eq(count, 100)

                # Trigger a grow/bump in the middle of a scan; make sure new nodes come after current scan position
                iter = slab.scanByRange(b'', db=foo)
                for _ in range(50):
                    next(iter)

                iterback = slab.scanByRangeBack(b'ffffffffffffffffffffffffffffffff', db=foo)
                for _ in range(50):
                    next(iterback)

                multikey = b'\xff\xff\xff\xfe' + s_common.guid(2000).encode('utf8')
                mapsize = slab.mapsize
                count = 0

                # Write until we grow
                while mapsize == slab.mapsize:
                    count += 1
                    rv = slab.put(multikey, s_common.guid(count + 100000).encode('utf8') + byts, dupdata=True, db=foo)
                    self.true(rv)

                self.eq(50 + count, sum(1 for _ in iter))
                self.eq(50, sum(1 for _ in iterback))

                self.true(os.path.isfile(slab.optspath))

                # Trigger a grow/bump in the middle of a dup scan
                iter = slab.scanByDups(multikey, db=foo)
                next(iter)

                iter2 = slab.scanByFull(db=foo2)
                next(iter2)

                iterback = slab.scanByDupsBack(multikey, db=foo)
                next(iterback)

                iterback2 = slab.scanByFullBack(db=foo2)
                next(iterback2)

                multikey = b'\xff\xff\xff\xff' + s_common.guid(i + 150000).encode('utf8')
                for i in range(200):
                    slab.put(multikey, s_common.guid(i + 200000).encode('utf8') + byts, dupdata=True, db=foo)

                self.eq(count - 1, sum(1 for _ in iter))
                self.eq(99, sum(1 for _ in iter2))

                self.eq(count - 1, sum(1 for _ in iterback))
                self.eq(99, sum(1 for _ in iterback2))

            # lets ensure our mapsize / growsize persisted, and make sure readonly works
            async with await s_lmdbslab.Slab.anit(path, map_size=100000, readonly=True) as newdb:

                self.eq(10000, newdb.growsize)
                foo = newdb.initdb('foo', dupsort=True)
                for _, _ in newdb.scanByRange(b'', db=foo):
                    count += 1
                self.gt(count, 200)

                # Make sure readonly is really readonly
                self.raises(s_exc.IsReadOnly, newdb.dropdb, 'foo')
                self.raises(s_exc.IsReadOnly, newdb.put, b'1234', b'3456')
                self.raises(s_exc.IsReadOnly, newdb.replace, b'1234', b'3456')
                self.raises(s_exc.IsReadOnly, newdb.pop, b'1234')
                self.raises(s_exc.IsReadOnly, newdb.delete, b'1234')
                self.raises(s_exc.IsReadOnly, newdb.putmulti, ((b'1234', b'3456'),))

                # While we have the DB open in readonly, have another process write a bunch of data to cause the
                # map size to be increased

                ctx = multiprocessing.get_context('spawn')
                proc = ctx.Process(target=_writeproc, args=(path, ))
                proc.start()
                proc.join()

                # Now trigger a remap for me
                newdb.get(multikey, db=foo)

    async def test_lmdbslab_grow_putmulti(self):
        '''
        Test for a regression where putmulti's across a grow could corrupt the database

        Test for a regression where a generator being passed into a putmulti would result in a partial write
        '''
        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')
            data = [i.to_bytes(4, 'little') for i in range(1000)]

            async with await s_lmdbslab.Slab.anit(path, map_size=10000) as slab:
                # A putmulti across a grow
                before_mapsize = slab.mapsize
                kvpairs = [(x, x) for x in data]
                retn = slab.putmulti(kvpairs)
                self.eq(retn, (1000, 1000))

                after_mapsize1 = slab.mapsize
                self.gt(after_mapsize1, before_mapsize)

                # A putmulti across a grow with a generator passed in
                kvpairs = ((b' ' + x, x) for x in data)
                retn = slab.putmulti(kvpairs)
                self.eq(retn, (1000, 1000))
                after_mapsize2 = slab.mapsize
                self.gt(after_mapsize2, after_mapsize1)

    async def test_lmdbslab_iternext_repeat_regression(self):
        '''
        Test for a scan being bumped in an iternext where the cursor is in the middle of a list of values with the same
        key
        '''

        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')
            my_maxsize = 500000

            async with await s_lmdbslab.Slab.anit(path, map_size=100000, growsize=50000, maxsize=my_maxsize) as slab:
                foo = slab.initdb('foo', dupsort=True)

                key = b'foo'
                for i in range(100):
                    slab.put(key, s_common.guid(i).encode('utf8'), db=foo)

                count = 0
                for _, _ in slab.scanByRange(b'', db=foo):
                    count += 1
                self.eq(count, 100)

                # Partially read through scan
                iter = slab.scanByRange(lmin=key, lmax=key, db=foo)
                for _ in range(60):
                    next(iter)

                # Trigger a bump by writing a bunch; make sure we're not writing into the middle of the scan
                multikey = b'\xff\xff\xff\xff' + s_common.guid(200).encode('utf8')
                mapsize = slab.mapsize
                count = 0
                while mapsize == slab.mapsize:
                    count += 1
                    slab.put(multikey, s_common.guid(count).encode('utf8') + b'0' * 256, dupdata=True, db=foo)

                # we wrote 100, read 60.  We should read only another 40
                self.len(40, list(iter))

    async def test_slab_guid_stor(self):

        with self.getTestDir() as dirn:
            path = os.path.join(dirn, 'slab.lmdb')
            async with await s_lmdbslab.Slab.anit(path) as slab:
                guidstor = s_lmdbslab.GuidStor(slab, 'guids')

                info0 = guidstor.gen('aaaa')
                info0.set('hehe', 20)
                self.eq(20, info0.get('hehe'))
                self.none(info0.get('haha'))

                info0.set('woot', {'woot': 1})
                self.eq((('hehe', 20), ('woot', {'woot': 1})), info0.items())

                self.eq({'woot': 1}, info0.get('woot'))
                self.eq({'woot': 1}, info0.pop('woot'))
                self.none(info0.get('woot'))
                self.none(info0.pop('woot'))
                self.true(info0.pop('woot', s_common.novalu) is s_common.novalu)

                # Sad path case
                self.raises(s_exc.NotMsgpackSafe, info0.set, 'newp', {1, 2, 3})

            async with await s_lmdbslab.Slab.anit(path) as slab:
                guidstor = s_lmdbslab.GuidStor(slab, 'guids')
                info1 = guidstor.gen('aaaa')
                self.eq(20, info1.get('hehe'))
                self.none(info1.pop('woot'))
                self.len(1, info1.items())
                self.eq((('hehe', 20), ), info1.items())

    async def test_slab_initdb_grow(self):
        self.thisHostMust(platform='linux')

        with self.getTestDir() as dirn:
            path = os.path.join(dirn, 'slab.lmdb')
            async with await s_lmdbslab.Slab.anit(path, map_size=1024, lockmemory=True) as slab:
                self.true(await asyncio.wait_for(slab.lockdoneevent.wait(), 8))
                mapcount = getFileMapCount('slab.lmdb/data.mdb')
                self.eq(1, mapcount)

                mapsize = slab.mapsize
                [slab.initdb(str(i)) for i in range(10)]
                self.gt(slab.mapsize, mapsize)

                # Make sure there is still only one map
                self.true(await asyncio.wait_for(slab.lockdoneevent.wait(), 8))

                mapcount = getFileMapCount('slab.lmdb/data.mdb')
                self.eq(1, mapcount)

    def test_slab_math(self):
        self.eq(s_lmdbslab._mapsizeround(100), 128)
        self.eq(s_lmdbslab._mapsizeround(s_const.mebibyte), s_const.mebibyte)
        self.eq(s_lmdbslab._mapsizeround(s_const.mebibyte + 1), 2 * s_const.mebibyte)
        self.eq(s_lmdbslab._mapsizeround(65 * s_const.gibibyte), 100 * s_const.gibibyte)
        self.eq(s_lmdbslab._mapsizeround(472 * s_const.gibibyte), 500 * s_const.gibibyte)
        self.eq(s_lmdbslab._mapsizeround(1000 * s_const.gibibyte), 1000 * s_const.gibibyte)

    async def test_slab_infinite_loop(self):
        '''
        Trigger a map full when replaying the log from a prior map full.
        '''
        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')
            byts = b'\x00' * 256

            count = 0
            async with await s_lmdbslab.Slab.anit(path, map_size=32000, growsize=5000, lockmemory=True) as slab:
                foo = slab.initdb('foo')
                slab.put(b'abcd', s_common.guid(count).encode('utf8') + byts, db=foo)
                await asyncio.sleep(1.1)
                count += 1
                slab.put(b'abcd', s_common.guid(count).encode('utf8') + byts, db=foo)

            # If we got here we're good
            self.true(True)

    async def test_slab_mapfull_runsyncloop(self):
        '''
        forcecommit in runSyncLoop can very occasionally trigger a mapfull
        '''
        with patch('synapse.lib.lmdbslab.Slab.DEFAULT_MAPSIZE', s_const.mebibyte), \
                patch('synapse.lib.lmdbslab.Slab.DEFAULT_GROWSIZE', 128 * s_const.kibibyte):
            batchsize = 4000
            numbatches = 2
            async with self.getTestCore() as core:
                before_mapsize = core.view.layers[0].layrslab.mapsize
                for i in range(numbatches):
                    async with await core.snap() as snap:
                        ips = ((('test:int', i * 1000000 + x), {'props': {'loc': 'us'}}) for x in range(batchsize))
                        await alist(snap.addNodes(ips))
                        # Wait for the syncloop to run
                        await asyncio.sleep(1.1)

                # Verify that it hit
                self.gt(core.view.layers[0].layrslab.mapsize, before_mapsize)

    async def test_slab_mapfull_drop(self):
        '''
        Test a mapfull in the middle of a dropdb
        '''
        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')
            data = [i.to_bytes(4, 'little') for i in range(400)]

            async with await s_lmdbslab.Slab.anit(path, map_size=32000, growsize=5000) as slab:
                slab.initdb('foo')
                kvpairs = [(x, x) for x in data]
                slab.putmulti(kvpairs)
                slab.forcecommit()
                before_mapsize = slab.mapsize
                slab.dropdb('foo')
                self.false(slab.dbexists('foo'))
                self.gt(slab.mapsize, before_mapsize)

    @staticmethod
    def make_slab(path):
        '''
        Multiprocessing target for expanding an existing slab
        '''
        async def workloop():
            s_glob.iAmLoop()
            data = [i.to_bytes(4, 'little') for i in range(400)]
            async with await s_lmdbslab.Slab.anit(path, map_size=32000, growsize=5000) as slab:
                slab.initdb('foo')
                kvpairs = [(x, x) for x in data]
                slab.putmulti(kvpairs)
                slab.forcecommit()

        asyncio.run(workloop())

    async def test_slab_mapfull_initdb(self):
        '''
        Test a mapfull in the middle of an initdb
        '''
        mpctx = multiprocessing.get_context('spawn')
        with self.getTestDir() as dirn:
            path = os.path.join(dirn, 'test.lmdb')
            async with await s_lmdbslab.Slab.anit(path, map_size=32000) as slab:
                pass
            async with await s_lmdbslab.Slab.anit(path, map_size=32000, readonly=True) as slab:

                proc = mpctx.Process(target=self.make_slab, args=(path,))
                proc.start()
                proc.join(10)
                self.nn(proc.exitcode)
                slab.initdb('foo')
                self.true(True)

    async def test_lmdb_multiqueue(self):

        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')

            async with await s_lmdbslab.Slab.anit(path) as slab:

                mque = await slab.getMultiQueue('test')

                self.false(mque.exists('woot'))

                with self.raises(s_exc.NoSuchName):
                    await mque.rem('woot')

                with self.raises(s_exc.NoSuchName):
                    await mque.get('woot', 0)

                with self.raises(s_exc.NoSuchName):
                    await mque.put('woot', 'lulz')

                with self.raises(s_exc.NoSuchName):
                    mque.status('woot')

                with self.raises(s_exc.NoSuchName):
                    await mque.cull('woot', -1)

                with self.raises(s_exc.NoSuchName):
                    await mque.dele('woot', 1, 1)

                with self.raises(s_exc.NoSuchName):
                    await mque.sets('woot', 1, ('lols',))

                await mque.add('woot', {'some': 'info'})
                await self.asyncraises(s_exc.DupName, mque.add('woot', {}))

                self.true(mque.exists('woot'))

                self.eq(0, await mque.put('woot', 'hehe'))
                self.eq(1, await mque.put('woot', 'haha'))
                self.eq(2, await mque.put('woot', 'hoho'))

                self.eq(3, mque.size('woot'))

                self.eq(3, await mque.put('woot', 'lol', reqid='foo'))
                self.eq(4, await mque.put('woot', 'lol', reqid='foo'))
                self.eq(4, await mque.put('woot', 'lol', reqid='foo'))

                self.eq(4, await mque.puts('woot', ('lol2', 'lol3'), reqid='foo2'))
                self.eq(6, await mque.puts('woot', ('lol2', 'lol3'), reqid='foo2'))
                self.eq(6, await mque.puts('woot', ('lol2', 'lol3'), reqid='foo2'))

                self.eq((0, 'hehe'), await mque.get('woot', 0))
                self.eq((1, 'haha'), await mque.get('woot', 1))
                self.eq((1, 'haha'), await mque.get('woot', 0))

                self.eq((-1, None), await mque.get('woot', 1000, cull=False))

                self.eq(5, mque.size('woot'))

                status = mque.list()
                self.len(1, status)
                self.eq(status[0], {'name': 'woot',
                                    'meta': {'some': 'info'},
                                    'size': 5,
                                    'offs': 6,
                                    })

                await mque.cull('woot', -1)
                self.eq(mque.status('woot'), status[0])

            async with await s_lmdbslab.Slab.anit(path) as slab:

                mque = await slab.getMultiQueue('test')

                self.eq(5, mque.size('woot'))
                self.eq(6, mque.offset('woot'))

                self.eq(((1, 'haha'), ), [x async for x in mque.gets('woot', 0, size=1)])

                correct = ((1, 'haha'), (2, 'hoho'), (3, 'lol'), (4, 'lol2'), (5, 'lol3'))
                self.eq(correct, [x async for x in mque.gets('woot', 0)])

                data = []
                evnt = asyncio.Event()

                async def getswait():
                    async for item in mque.gets('woot', 0, wait=True):

                        if item[1] is None:
                            break

                        data.append(item)

                        if item[1] == 'hoho':
                            evnt.set()

                task = slab.schedCoro(getswait())

                await asyncio.wait_for(evnt.wait(), 5)

                self.eq(data, correct)

                await mque.put('woot', 'lulz')
                await mque.put('woot', None)

                await asyncio.wait_for(task, 2)

                self.eq(data, (*correct, (6, 'lulz')))

                self.true(mque.exists('woot'))

                self.eq((2, 'hoho'), await mque.get('woot', 2))

                await mque.put('woot', 'huhu')

                await mque.rem('woot')

                self.false(mque.exists('woot'))

                await mque.add('woot', {'some': 'info'})
                self.eq(0, await mque.put('woot', 'hehe'))
                self.eq(1, await mque.put('woot', 'haha'))
                self.eq(2, await mque.put('woot', 'hoho'))

                self.eq(3, mque.size('woot'))
                self.eq(3, mque.offset('woot'))

                # Replace one item in the queue
                await mque.sets('woot', 1, ('lol',))
                self.eq(3, mque.size('woot'))
                self.eq(3, mque.offset('woot'))

                correct = ((0, 'hehe'), (1, 'lol'), (2, 'hoho'))
                self.eq(correct, [x async for x in mque.gets('woot', 0)])

                # Replace multiple items in the queue
                await mque.sets('woot', 1, ('lol2', 'lol3'))
                self.eq(3, mque.size('woot'))
                self.eq(3, mque.offset('woot'))

                correct = ((0, 'hehe'), (1, 'lol2'), (2, 'lol3'))
                self.eq(correct, [x async for x in mque.gets('woot', 0)])

                # Replace items going past the end of the current queue
                await mque.sets('woot', 2, ('lol4', 'lol5', 'lol6'))
                self.eq(5, mque.size('woot'))
                self.eq(5, mque.offset('woot'))

                correct = ((0, 'hehe'), (1, 'lol2'), (2, 'lol4'), (3, 'lol5'), (4, 'lol6'))
                self.eq(correct, [x async for x in mque.gets('woot', 0)])

                # Delete from the middle of the queue
                await mque.dele('woot', 1, 3)
                self.eq(2, mque.size('woot'))
                self.eq(5, mque.offset('woot'))

                correct = ((0, 'hehe'), (4, 'lol6'))
                self.eq(correct, [x async for x in mque.gets('woot', 0)])

                # Add items in the gap we created
                await mque.sets('woot', 2, ('lol7', 'lol8'))
                self.eq(4, mque.size('woot'))
                self.eq(5, mque.offset('woot'))

                correct = ((0, 'hehe'), (2, 'lol7'), (3, 'lol8'), (4, 'lol6'))
                self.eq(correct, [x async for x in mque.gets('woot', 0)])

                # Delete a partially empty range
                await mque.dele('woot', 3, 10)
                self.eq(2, mque.size('woot'))
                self.eq(5, mque.offset('woot'))

                correct = ((0, 'hehe'), (2, 'lol7'))
                self.eq(correct, [x async for x in mque.gets('woot', 0)])

                # Delete a completely empty range
                await mque.dele('woot', 100, 150)
                self.eq(2, mque.size('woot'))
                self.eq(5, mque.offset('woot'))

                correct = ((0, 'hehe'), (2, 'lol7'))
                self.eq(correct, [x async for x in mque.gets('woot', 0)])

                # Set items past the end of the current queue
                await mque.sets('woot', 200, ('lol9', 'lol0'))
                self.eq(4, mque.size('woot'))
                self.eq(202, mque.offset('woot'))

                correct = ((0, 'hehe'), (2, 'lol7'), (200, 'lol9'), (201, 'lol0'))
                self.eq(correct, [x async for x in mque.gets('woot', 0)])

                # Adding items past the current end of queue should wake waiters
                data = []

                async def getswait():
                    async for item in mque.gets('woot', 0, wait=True):

                        data.append(item)

                        if item[1] == 'lol0':
                            break

                task = slab.schedCoro(getswait())

                await mque.sets('woot', 201, ('lol9', 'lol0'))
                await asyncio.wait_for(task, 2)

                self.eq(5, mque.size('woot'))
                self.eq(203, mque.offset('woot'))

                correct = ((0, 'hehe'), (2, 'lol7'), (200, 'lol9'), (201, 'lol9'), (202, 'lol0'))
                self.eq(data, correct)

                # Invalid offsets that won't do anything
                await mque.dele('woot', -1, 20)
                await mque.dele('woot', -5, -1)
                await mque.dele('woot', 5, 1)
                await mque.dele('woot', 5, -1)
                await mque.sets('woot', -1, ('lolz', 'lol'))

                self.eq(5, mque.size('woot'))
                self.eq(203, mque.offset('woot'))

                correct = ((0, 'hehe'), (2, 'lol7'), (200, 'lol9'), (201, 'lol9'), (202, 'lol0'))
                self.eq(correct, [x async for x in mque.gets('woot', 0)])

    async def test_slababrv(self):
        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')

            async with await s_lmdbslab.Slab.anit(path) as slab:
                abrv = s_lmdbslab.SlabAbrv(slab, 'test')

                valu = abrv.setBytsToAbrv('hehe'.encode())
                self.eq(valu, b'\x00\x00\x00\x00\x00\x00\x00\x00')
                valu = abrv.setBytsToAbrv('haha'.encode())
                self.eq(valu, b'\x00\x00\x00\x00\x00\x00\x00\x01')

                name = abrv.abrvToByts(b'\x00\x00\x00\x00\x00\x00\x00\x01')
                self.eq(name, b'haha')

                self.raises(s_exc.NoSuchAbrv, abrv.abrvToByts, b'\x00\x00\x00\x00\x00\x00\x00\x02')

            # And persistence
            async with await s_lmdbslab.Slab.anit(path) as slab:
                abrv = s_lmdbslab.SlabAbrv(slab, 'test')
                # recall first
                name = abrv.abrvToByts(b'\x00\x00\x00\x00\x00\x00\x00\x00')
                self.eq(name, b'hehe')

                name = abrv.abrvToByts(b'\x00\x00\x00\x00\x00\x00\x00\x01')
                self.eq(name, b'haha')
                # Remaking them makes the values we already had
                valu = abrv.nameToAbrv('hehe')
                self.eq(valu, b'\x00\x00\x00\x00\x00\x00\x00\x00')

                valu = abrv.nameToAbrv('haha')
                self.eq(valu, b'\x00\x00\x00\x00\x00\x00\x00\x01')

                self.eq('haha', abrv.abrvToName(b'\x00\x00\x00\x00\x00\x00\x00\x01'))

                # And we still have no valu for 02
                self.raises(s_exc.NoSuchAbrv, abrv.abrvToByts, b'\x00\x00\x00\x00\x00\x00\x00\x02')

                # And we don't overwrite existing values on restart
                valu = abrv.setBytsToAbrv('hoho'.encode())
                self.eq(valu, b'\x00\x00\x00\x00\x00\x00\x00\x02')

                valu = abrv.nameToAbrv('haha')
                self.eq(valu, b'\x00\x00\x00\x00\x00\x00\x00\x01')

    async def test_lmdbslab_hotkeyval(self):
        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')

            async with await s_lmdbslab.Slab.anit(path, map_size=1000000) as slab, \
                    await s_lmdbslab.HotKeyVal.anit(slab, 'counts') as ctr:
                self.eq(None, ctr.get('foo'))
                self.eq({}, ctr.pack())
                ctr.set('foo', 1)
                ctr.set('bar', {'val': 42})
                self.eq({'foo': 1, 'bar': {'val': 42}}, ctr.pack())
                ctr.set('baz', 42)
                ctr.delete('baz')
                self.eq(None, ctr.get('baz'))

            async with await s_lmdbslab.Slab.anit(path, map_size=1000000) as slab, \
                    await s_lmdbslab.HotKeyVal.anit(slab, 'counts') as ctr:
                self.eq({'foo': 1, 'bar': {'val': 42}}, ctr.pack())
                self.eq({'val': 42}, ctr.get('bar'))

    async def test_lmdbslab_hotcount(self):

        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')

            async with await s_lmdbslab.Slab.anit(path, map_size=1000000, lockmemory=True) as slab, \
                    await s_lmdbslab.HotCount.anit(slab, 'counts') as ctr:
                self.eq(0, ctr.get('foo'))
                self.eq({}, ctr.pack())
                ctr.inc('foo')
                self.eq({'foo': 1}, ctr.pack())
                self.eq(1, ctr.get('foo'))
                ctr.set('bar', 42)
                self.eq({'foo': 1, 'bar': 42}, ctr.pack())
                ctr.sync()
                self.eq({'foo': 1, 'bar': 42}, ctr.pack())

                ctr.inc('foo')
                ctr.inc('foo')
                ctr.set('bar', 37)
                ctr.set('bar', -42)
                ctr.sync()

                cache = []
                for lkey, lval in slab.scanByFull(db='counts'):
                    cache.append((lkey, s_common.int64un(lval)))

                self.len(1, [k for k, v in cache if k == b'foo'])
                self.len(1, [k for k, v in cache if k == b'bar'])

    async def test_lmdbslab_doubleopen(self):

        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')
            async with await s_lmdbslab.Slab.anit(path) as slab:
                foo = slab.initdb('foo')
                slab.put(b'\x00\x01', b'hehe', db=foo)

            # Can close and re-open fine
            async with await s_lmdbslab.Slab.anit(path) as slab:
                foo = slab.initdb('foo')
                self.eq(b'hehe', slab.get(b'\x00\x01', db=foo))

                # Can't re-open while already open
                await self.asyncraises(s_exc.SlabAlreadyOpen, s_lmdbslab.Slab.anit(path))

    async def test_lmdbslab_copyslab(self):

        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')
            copypath = os.path.join(dirn, 'copy.lmdb')

            async with await s_lmdbslab.Slab.anit(path) as slab:
                foo = slab.initdb('foo')
                slab.put(b'\x00\x01', b'hehe', db=foo)

                await slab.copyslab(copypath)

                self.true(pathlib.Path(copypath).with_suffix('.opts.yaml').exists())

                async with await s_lmdbslab.Slab.anit(copypath) as slabcopy:
                    foo = slabcopy.initdb('foo')
                    self.eq(b'hehe', slabcopy.get(b'\x00\x01', db=foo))

                await self.asyncraises(s_exc.DataAlreadyExists, slab.copyslab(copypath))

    async def test_lmdbslab_statinfo(self):

        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')

            async with await s_lmdbslab.Slab.anit(path) as slab:

                foo = slab.initdb('foo')

                slab.put(b'\x00\x01', b'hehe', db=foo)
                slab.put(b'\x00\x02', b'haha', db=foo)
                await slab.sync()

                stats = slab.statinfo()

                self.false(stats['locking_memory'])
                self.false(stats['prefaulting'])

                commitstats = stats['commitstats']
                self.len(2, commitstats)
                self.eq(2, commitstats[-1][1])

    async def test_lmdbslab_iter_and_delete(self):
        with self.getTestDir() as dirn:
            path = os.path.join(dirn, 'test.lmdb')
            async with await s_lmdbslab.Slab.anit(path, map_size=1000000, lockmemory=True) as slab:
                bar = slab.initdb('bar', dupsort=True)
                slab.put(b'\x00\x01', b'hehe', dupdata=True, db=bar)
                slab.put(b'\x00\x02', b'haha', dupdata=True, db=bar)
                scan = slab.scanByDups(b'\x00\x02', db=bar)
                self.eq((b'\x00\x02', b'haha'), next(scan))
                slab.delete(b'\x00\x01', b'hehe', db=bar)
                self.raises(StopIteration, next, scan)

    async def test_lmdbslab_hist(self):
        with self.getTestDir() as dirn:
            path = os.path.join(dirn, 'test.lmdb')
            async with await s_lmdbslab.Slab.anit(path, map_size=1000000) as slab:
                now = s_common.now()
                hist = s_lmdbslab.Hist(slab, 'history')
                hist.add('foo')
                await asyncio.sleep(0.1)
                hist.add('bar')
                then = s_common.now()

                items = []
                for item in hist.carve(now, then):
                    items.append(item)
                self.len(2, items)
                self.eq([item[1] for item in items], ['foo', 'bar'])

                # Carve from an arbitrary point forward to the end
                tick = items[1][0]
                await asyncio.sleep(0.1)
                hist.add('baz')
                items = []
                for item in hist.carve(tick):
                    items.append(item)
                self.len(2, items)
                self.eq([item[1] for item in items], ['bar', 'baz'])

                # Add a item at given tick and carve it
                hist.add('timewarp', tick=now - 10)

                items = []
                for item in hist.carve(now - 15, then):
                    items.append(item)
                self.len(3, items)
                self.eq([item[1] for item in items], ['timewarp', 'foo', 'bar'])

                # boundary conditions

                # Minimum values
                hist.add('bot', tick=0)
                with self.raises(OverflowError):
                    hist.add('bot', tick=-1)

                # Maximum value we can store
                hist.add('eot', tick=(2 * 9223372036854775807) + 1)
                with self.raises(OverflowError):
                    hist.add('eot', tick=(2 * 9223372036854775807) + 2)

                # Tablescan
                items = []
                for item in hist.carve(0):
                    items.append(item[1])
                self.len(6, items)
                self.eq(items, ['bot', 'timewarp', 'foo', 'bar', 'baz', 'eot'])


class LmdbSlabMemLockTest(s_t_utils.SynTest):

    async def test_lmdbslabmemlock(self):
        self.thisHostMust(hasmemlocking=True)

        beforelockmem = s_thisplat.getCurrentLockedMemory()

        with self.getTestDir() as dirn:

            path = os.path.join(dirn, 'test.lmdb')
            async with await s_lmdbslab.Slab.anit(path, map_size=1000000, lockmemory=True) as lmdbslab:

                self.true(await asyncio.wait_for(lmdbslab.lockdoneevent.wait(), 8))
                lockmem = s_thisplat.getCurrentLockedMemory()
                self.ge(lockmem - beforelockmem, 4000)

    async def test_multiple_grow(self):
        '''
        Trigger multiple grow events rapidly and ensure memlock thread survives.
        '''
        self.thisHostMust(hasmemlocking=True)

        with self.getTestDir() as dirn:

            count = 0
            byts = b'\x00' * 1024
            path = os.path.join(dirn, 'test.lmdb')
            mapsize = 10 * 1024 * 1024
            async with await s_lmdbslab.Slab.anit(path, map_size=mapsize, growsize=5000, lockmemory=True) as slab:
                foo = slab.initdb('foo')
                while count < 8000:
                    count += 1
                    slab.put(s_common.guid(count).encode('utf8'), s_common.guid(count).encode('utf8') + byts, db=foo)

                self.true(await asyncio.wait_for(slab.lockdoneevent.wait(), 8))

                lockmem = s_thisplat.getCurrentLockedMemory()

                # TODO: make this test reliable
                self.ge(lockmem, 0)

    async def test_math(self):
        self.eq(16, s_lmdbslab._florpo2(16))
        self.eq(16, s_lmdbslab._florpo2(17))
        self.eq(16, s_lmdbslab._florpo2(31))

        self.eq(16, s_lmdbslab._ceilpo2(16))
        self.eq(16, s_lmdbslab._ceilpo2(15))
        self.eq(16, s_lmdbslab._ceilpo2(9))

        self.eq(4, s_lmdbslab._roundup(4, 2))
        self.eq(4, s_lmdbslab._roundup(3, 2))

def _writeproc(path):

    async def lotsofwrites(path):
        byts = b'\x00' * 256
        os.remove(pathlib.Path(path).with_suffix('.opts.yaml'))
        async with await s_lmdbslab.Slab.anit(path, map_size=100000) as slab:
            foo = slab.initdb('foo', dupsort=True)
            mapsize = slab.mapsize
            count = 0
            while mapsize == slab.mapsize:
                count += 1
                slab.put(b'abcd', s_common.guid(count).encode('utf8') + byts, dupdata=True, db=foo)
    asyncio.run(lotsofwrites(path))



import socket
import asyncio

import synapse.common as s_common

import synapse.lib.coro as s_coro
import synapse.lib.link as s_link

import synapse.tests.utils as s_test

class LinkTest(s_test.SynTest):

    async def test_link_raw(self):

        async def onlink(link):
            self.eq(b'vis', await link.recvsize(3))
            self.eq(b'i', await link.recv(1))
            await link.send(b'vert')
            await link.fini()

        serv = await s_link.listen('127.0.0.1', 0, onlink)
        host, port = serv.sockets[0].getsockname()

        link = await s_link.connect(host, port)

        await link.send(b'visi')
        self.eq(b'vert', await link.recvsize(4))
        self.none(await link.recvsize(1))

    async def test_link_tx_sadpath(self):

        evt = asyncio.Event()

        async def onlink(link):
            msg0 = await link.rx()
            self.eq(('what', {'k': 1}), msg0)
            link.onfini(evt.set)
            await link.fini()

        serv = await s_link.listen('127.0.0.1', 0, onlink)
        host, port = serv.sockets[0].getsockname()
        link = await s_link.connect(host, port)
        await link.tx(('what', {'k': 1}))
        self.true(await s_coro.event_wait(evt, 6))
        # Why does this first TX post fini on the server link work,
        # but the second one fails?
        await link.tx(('me', {'k': 2}))
        await self.asyncraises(ConnectionError, link.tx(('worry?', {'k': 3})))

    async def test_link_rx_sadpath(self):

        junk = 'a9d0cdafef705b9864bd'
        # random sequence of data which causes an error
        # to be thrown when unpacking data via msgpack.
        junk = s_common.uhex(junk)

        evt = asyncio.Event()

        async def onlink(link):
            await link.tx(('what', {'k': 1}))
            self.true(await s_coro.event_wait(evt, 6))
            # Send purposely bad data through the link
            await link.send(junk)
            await link.fini()

        serv = await s_link.listen('127.0.0.1', 0, onlink)
        host, port = serv.sockets[0].getsockname()
        link = await s_link.connect(host, port)
        msg0 = await link.rx()
        self.eq(msg0, ('what', {'k': 1}))
        evt.set()
        await asyncio.sleep(0)
        with self.getAsyncLoggerStream('synapse.lib.link', 'rx error') as stream:
            msg1 = await link.rx()
            self.true(await stream.wait(6))
        self.none(msg1)

    async def test_link_file(self):

        link0, file0 = await s_link.linkfile('rb')

        def reader(fd):
            byts = fd.read()
            fd.close()
            return byts

        coro = s_coro.executor(reader, file0)

        await link0.send(b'asdf')
        await link0.send(b'qwer')

        await link0.fini()

        self.eq(b'asdfqwer', await coro)

        link1, file1 = await s_link.linkfile('wb')

        def writer(fd):
            fd.write(b'asdf')
            fd.write(b'qwer')
            fd.close()

        coro = s_coro.executor(writer, file1)

        byts = b''

        while True:
            x = await link1.recv(1000000)
            if not x:
                break
            byts += x

        await coro

        self.eq(b'asdfqwer', byts)

    async def test_linksock(self):
        link0, sock0 = await s_link.linksock()
        self.isinstance(link0, s_link.Link)
        self.isinstance(sock0, socket.socket)

        def reader(sock):
            buf = b''
            while True:
                byts = sock.recv(1024)
                if not byts:
                    break
                buf += byts
            return buf

        coro = s_coro.executor(reader, sock0)

        await link0.send(b'part1')
        await link0.send(b'qwer')

        await link0.fini()

        self.eq(b'part1qwer', await coro)
        sock0.close()

        link1, sock1 = await s_link.linksock()

        def writer(sock):
            sock.sendall(b'part2')
            sock.sendall(b'qwer')
            sock.shutdown(socket.SHUT_WR)

        coro = s_coro.executor(writer, sock1)
        await coro

        self.eq(b'part2qwer', await link1.recvsize(9))

        await link1.fini()
        sock1.close()

    async def test_link_fromspawn(self):

        link0, sock0 = await s_link.linksock()

        info = await link0.getSpawnInfo()
        link1 = await s_link.fromspawn(info)

        await link1.send(b'V')
        self.eq(sock0.recv(1), b'V')

        sock0.close()

        await link0.fini()
        await link1.fini()



import os
import json
import shlex
import pprint
import asyncio
import tempfile
import functools
import subprocess

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.cmd as s_cmd
import synapse.lib.cli as s_cli

ListHelp = '''
Lists all the keys underneath a particular key in the hive.

Syntax:
    hive ls|list [path]

Notes:
    If path is not specified, the root is listed.
'''

GetHelp = '''
Display or save to file the contents of a key in the hive.

Syntax:
    hive get [--file] [--json] {path}
'''

DelHelp = '''
Deletes a key in the cell's hive.

Syntax:
    hive rm|del {path}

Notes:
    Delete will recursively delete all subkeys underneath path if they exist.
'''

EditHelp = '''
Edits or creates a key in the cell's hive.

Syntax:
    hive edit|mod {path} [--string] ({value} | --editor | -f {filename})

Notes:
    One may specify the value directly on the command line, from a file, or use an editor.  For the --editor option,
    the environment variable VISUAL or EDITOR must be set.
'''

class HiveCmd(s_cli.Cmd):
    '''
Manipulates values in a cell's Hive.

A Hive is a hierarchy persistent storage mechanism typically used for configuration data.
'''
    _cmd_name = 'hive'

    _cmd_syntax = (
        ('line', {'type': 'glob'}),  # type: ignore
    )

    def _make_argparser(self):

        parser = s_cmd.Parser(prog='hive', outp=self, description=self.__doc__)

        subparsers = parser.add_subparsers(title='subcommands', required=True, dest='cmd',
                                           parser_class=functools.partial(s_cmd.Parser, outp=self))

        parser_ls = subparsers.add_parser('list', aliases=['ls'], help="List entries in the hive", usage=ListHelp)
        parser_ls.add_argument('path', nargs='?', help='Hive path')

        parser_get = subparsers.add_parser('get', help="Get any entry in the hive", usage=GetHelp)
        parser_get.add_argument('path', help='Hive path')
        parser_get.add_argument('-f', '--file', default=False, action='store',
                                help='Save the data to a file.')
        parser_get.add_argument('--json', default=False, action='store_true', help='Emit output as json')

        parser_rm = subparsers.add_parser('del', aliases=['rm'], help='Delete a key in the hive', usage=DelHelp)
        parser_rm.add_argument('path', help='Hive path')

        parser_edit = subparsers.add_parser('edit', aliases=['mod'], help='Sets/creates a key', usage=EditHelp)
        parser_edit.add_argument('--string', action='store_true', help="Edit value as a single string")
        parser_edit.add_argument('path', help='Hive path')
        group = parser_edit.add_mutually_exclusive_group(required=True)
        group.add_argument('value', nargs='?', help='Value to set')
        group.add_argument('--editor', default=False, action='store_true',
                           help='Opens an editor to set the value')
        group.add_argument('--file', '-f', help='Copies the contents of the file to the path')

        return parser

    async def runCmdOpts(self, opts):
        line = opts.get('line')
        if line is None:
            self.printf(self.__doc__)
            return

        core = self.getCmdItem()

        try:
            opts = self._make_argparser().parse_args(shlex.split(line))
        except s_exc.ParserExit:
            return

        handlers = {
            'list': self._handle_ls,
            'ls': self._handle_ls,
            'del': self._handle_rm,
            'rm': self._handle_rm,
            'get': self._handle_get,
            'edit': self._handle_edit,
            'mod': self._handle_edit,
        }
        await handlers[opts.cmd](core, opts)

    @staticmethod
    def parsepath(path):
        ''' Turn a slash-delimited path into a list that hive takes '''
        return path.split('/')

    async def _handle_ls(self, core, opts):
        path = self.parsepath(opts.path) if opts.path is not None else None
        keys = await core.listHiveKey(path=path)
        if keys is None:
            self.printf('Path not found')
            return
        for key in keys:
            self.printf(key)

    async def _handle_get(self, core, opts):
        path = self.parsepath(opts.path)

        valu = await core.getHiveKey(path)
        if valu is None:
            self.printf(f'{opts.path} not present')
            return

        if opts.json:
            prend = json.dumps(valu, indent=4, sort_keys=True)
            rend = prend.encode()
        elif isinstance(valu, str):
            rend = valu.encode()
            prend = valu
        elif isinstance(valu, bytes):
            rend = valu
            prend = pprint.pformat(valu)
        else:
            rend = json.dumps(valu, indent=4, sort_keys=True).encode()
            prend = pprint.pformat(valu)

        if opts.file:
            with s_common.genfile(opts.file) as fd:
                fd.truncate(0)
                fd.write(rend)
            self.printf(f'Saved the hive entry [{opts.path}] to {opts.file}')
            return

        self.printf(f'{opts.path}:\n{prend}')

    async def _handle_rm(self, core, opts):
        path = self.parsepath(opts.path)
        await core.popHiveKey(path)

    async def _handle_edit(self, core, opts):
        path = self.parsepath(opts.path)

        if opts.value is not None:
            if opts.value[0] not in '([{"':
                data = opts.value
            else:
                data = json.loads(opts.value)
            await core.setHiveKey(path, data)
            return
        elif opts.file is not None:
            with open(opts.file) as fh:
                s = fh.read()
                if len(s) == 0:
                    self.printf('Empty file.  Not writing key.')
                    return
                data = s if opts.string else json.loads(s)
                await core.setHiveKey(path, data)
                return

        editor = os.getenv('VISUAL', (os.getenv('EDITOR', None)))
        if editor is None or editor == '':
            self.printf('Environment variable VISUAL or EDITOR must be set for --editor')
            return
        tnam = None
        try:
            with tempfile.NamedTemporaryFile(mode='w', delete=False) as fh:
                old_valu = await core.getHiveKey(path)
                if old_valu is not None:
                    if opts.string:
                        if not isinstance(old_valu, str):
                            self.printf('Existing value is not a string, therefore not editable as a string')
                            return
                        data = old_valu
                    else:
                        try:
                            data = json.dumps(old_valu, indent=4, sort_keys=True)
                        except (ValueError, TypeError):
                            self.printf('Value is not JSON-encodable, therefore not editable.')
                            return
                    fh.write(data)
                tnam = fh.name
            while True:
                retn = subprocess.call(f'{editor} {tnam}', shell=True)
                if retn != 0:  # pragma: no cover
                    self.printf('Editor failed with non-zero code.  Aborting.')
                    return
                with open(tnam) as fh:
                    rawval = fh.read()
                    if len(rawval) == 0:  # pragma: no cover
                        self.printf('Empty file.  Not writing key.')
                        return
                    try:
                        valu = rawval if opts.string else json.loads(rawval)
                    except json.JSONDecodeError as e:  # pragma: no cover
                        self.printf(f'JSON decode failure: [{e}].  Reopening.')
                        await asyncio.sleep(1)
                        continue

                    # We lose the tuple/list distinction in the telepath round trip, so tuplify everything to compare
                    if (opts.string and valu == old_valu) or (not opts.string and s_common.tuplify(valu) == old_valu):
                        self.printf('Valu not changed.  Not writing key.')
                        return
                    await core.setHiveKey(path, valu)
                    break

        finally:
            if tnam is not None:
                os.unlink(tnam)



import hashlib

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.tests.utils as s_t_utils


BITS = 2048
HEXSTR_MODULUS = 'abbd407f417fe8d6632aae1c6d09b271416bef9244e61f7c7c2856ddfde3ecf93cd50b3eaea5c9b8cb9bfb5a317bf50925a' \
                 'b500a06247ec2f3294891a8e62c317ee648f933ec1bf760a9d7e9a5ea4706b2a2c3f6376079114ddcc7a15d3fecf001458f' \
                 '22f0551802a25ef95cf464aabeb0514ea3849583bc09022730c44a2ff5f893fc6885add69c103d75114dd2f11436f617fbf' \
                 'b0af2978802aabf35483bbfcc470d50d6afb4283c1d06d2bf27efe9d7c09f226895633a46c3d77173bf0db8634299462b5f' \
                 '29629ad3b0470c76ddfd331ed0207d4dbd5fd44a2f66ca5f802ac0130e4a4bb2c149b5baa7a373188823ee21fe2950a76c8' \
                 '18586919f7914453d'
HEXSTR_PUBLIC_EXPONENT = 0x10001
HEXSTR_PRIVATE_EXPONENT = '9db58a80120f3b2b7d1f998a231b8f916fa985f4456f2a24f0033f5a56a7b35b61e0a695e65dfab3c7ceb2f0ad' \
                          '968e7bdaeac9f29a97730ce5add8a5627c14c3532c7880d88c8f56099f8ed65275a4c9e2cb93b70c3d7c904677' \
                          '639fac7962c537f5bfaf2f12859d0dacb7c403ee59da0922715bba0a6f5202d7c653833e39715f04664c2396c4' \
                          '7bdf3f09f5486d8f6aea767ba011f1a5a10c8b57f079aea58abfd5e50ef20aa5e09b1082f6af98e806c9aeeb89' \
                          '4148a7d82cd6e1443c6115eb567fba0eacf5b7178518b8ba312da6ace22238d1ed19f3e703652576a6152ba60d' \
                          '4d4c6bc75b3ee7c8efeadee0c5ed7c14bf2930a6c4f13137becf38912f49c5'
HEXSTR_PRIVATE_PRIME_P = 'dee90ee63c12729a3fe7d38c581abf7e1c784ec0bd4bfdd1282286ea9996673942a24c7c98b31c6cd12db8ba96d' \
                         'a785c4392569d7bfc2be9d9907c3b7fbf40d31891642952a0e5a23dfbe721a746588df9a246ea4936a1958f66fd' \
                         '3a32c08008a0f6ed9b516fa869fb08a57ef31c0ec217f173e489a2f8f111e25c25c961c2b7'
HEXSTR_PRIVATE_PRIME_Q = 'c53b9c8dfb3dda04d16c7f779a02b3b8c7b44bf876dc88ad562778eafaded9ade882ccfb887761515a251c22476' \
                         '1bef7207fa489e398041787cfbd155f1034a207d517f06bc76a044262484f82f0c6a887f776b1dce837408999d8' \
                         '8dd33a96c7f80e23719e77a11075d337bf9cc47d7dbf98e341b81c23f165dd15ccfd2973ab'

TEST_MD5 = hashlib.md5(b'test').hexdigest()
TEST_SHA1 = hashlib.sha1(b'test').hexdigest()
TEST_SHA256 = hashlib.sha256(b'test').hexdigest()
TEST_SHA384 = hashlib.sha384(b'test').hexdigest()
TEST_SHA512 = hashlib.sha512(b'test').hexdigest()

class CryptoModelTest(s_t_utils.SynTest):

    async def test_model_crypto_currency(self):

        async with self.getTestCore() as core:

            nodes = await core.nodes('[ crypto:currency:client=(1.2.3.4, (btc, 1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2)) ]')
            self.len(1, nodes)

            nodes = await core.nodes('inet:client=1.2.3.4 -> crypto:currency:client -> crypto:currency:address')
            self.eq(nodes[0].get('coin'), 'btc')
            self.eq(nodes[0].get('iden'), '1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2')

            nodes = await core.nodes('''
                [
                    econ:acct:payment="*"
                        :from:coinaddr=(btc, 1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2)
                        :to:coinaddr=(btc, 1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2)
                ]
            ''')

            # these would explode if the model was wrong
            self.len(1, await core.nodes('crypto:currency:address [ :desc="woot woot" :contact="*" ] -> ps:contact'))
            self.len(1, await core.nodes('crypto:currency:address:iden=1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2'))
            self.len(1, await core.nodes('crypto:currency:address:coin=btc'))
            self.len(1, await core.nodes('crypto:currency:client:inetaddr=1.2.3.4'))

            opts = {'vars': {
                'input': hashlib.sha256(b'asdf').hexdigest(),
                'output': hashlib.sha256(b'qwer').hexdigest(),
            }}

            payors = await core.nodes('[ crypto:payment:input=* :transaction=(t1,) :address=(btc, 1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2) :value=30 ]')
            self.eq(payors[0].get('value'), '30')
            self.eq(payors[0].get('address'), ('btc', '1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2'))

            payees = await core.nodes('[ crypto:payment:output=* :transaction=(t1,) :address=(btc, 1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2) :value=30 ]')
            self.eq(payees[0].get('value'), '30')
            self.eq(payees[0].get('address'), ('btc', '1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2'))

            payor = payors[0].ndef[1]
            payee = payees[0].ndef[1]

            nodes = await core.nodes(f'''
                [
                    crypto:currency:transaction=(t1,)
                        :hash=0x01020304
                        :desc="Woot Woot"
                        :block=(BTC, 998877)
                        :success=1
                        :status:code=10
                        :status:message=success
                        :to = (btc, 1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2)
                        :from = (btc, 1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2)
                        :fee = 0.0001
                        :value = 30
                        :time = 20211031
                        :eth:gasused = 10
                        :eth:gaslimit = 20
                        :eth:gasprice = 0.001
                        :contract:input = $input
                        :contract:output = $output
                ]
            ''', opts=opts)
            self.len(1, nodes)
            node = nodes[0]
            self.eq(node.get('hash'), '01020304')
            self.eq(node.get('desc'), 'Woot Woot')
            self.eq(node.get('block'), ('btc', 998877))
            self.eq(node.get('block:coin'), 'btc')
            self.eq(node.get('block:offset'), 998877)
            self.eq(node.get('success'), True)
            self.eq(node.get('status:code'), 10)
            self.eq(node.get('status:message'), 'success')
            self.eq(node.get('to'), ('btc', '1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2'))
            self.eq(node.get('from'), ('btc', '1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2'))
            self.eq(node.get('fee'), '0.0001')
            self.eq(node.get('value'), '30')
            self.eq(node.get('time'), 1635638400000)
            self.eq(node.get('eth:gasused'), 10)
            self.eq(node.get('eth:gaslimit'), 20)
            self.eq(node.get('eth:gasprice'), '0.001')
            self.eq(node.get('contract:input'), 'sha256:f0e4c2f76c58916ec258f246851bea091d14d4247a2fc3e18694461b1816e13b')
            self.eq(node.get('contract:output'), 'sha256:f6f2ea8f45d8a057c9566a33f99474da2e5c6a6604d736121650e2730c6fb0a3')

            with self.raises(s_exc.IsDeprLocked):
                await node.set('inputs', (payor,))
            with self.raises(s_exc.IsDeprLocked):
                await node.set('outputs', (payee,))

            q = 'crypto:currency:transaction=(t1,) | tee { -> crypto:payment:input } { -> crypto:payment:output }'
            nodes = await core.nodes(q)
            self.eq({n.ndef[1] for n in nodes}, {payor, payee})

            nodes = await core.nodes('''
                [
                    crypto:currency:block=(btc, 12345)
                        :hash=0x01020304
                        :minedby = (btc, 1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2)
                        :time=20211130
                ]''')
            self.len(1, nodes)
            node = nodes[0]
            self.eq(node.get('coin'), 'btc')
            self.eq(node.get('offset'), 12345)
            self.eq(node.get('hash'), '01020304')
            self.eq(node.get('time'), 1638230400000)

            nodes = await core.nodes('''
                [
                    crypto:smart:contract=*
                        :transaction=*
                        :bytecode=$input
                        :address = (btc, 1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2)
                        :token:name=Foo
                        :token:symbol=Bar
                        :token:totalsupply=300
                ]''', opts=opts)
            self.len(1, nodes)
            node = nodes[0]
            self.nn(node.get('transaction'))
            self.eq(node.get('bytecode'), 'sha256:f0e4c2f76c58916ec258f246851bea091d14d4247a2fc3e18694461b1816e13b')
            self.eq(node.get('address'), ('btc', '1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2'))
            self.eq(node.get('token:name'), 'Foo')
            self.eq(node.get('token:symbol'), 'Bar')
            self.eq(node.get('token:totalsupply'), '300')

            nodes = await core.nodes('''
                [
                    crypto:smart:token=(2bdea834252a220b61aadf592cc0de66, 30)
                        :owner=eth/aaaa
                        :nft:url = https://coin.vertex.link/nfts/30
                        :nft:meta = $lib.dict(name=WootWoot)
                        :nft:meta:name = WootWoot
                        :nft:meta:description = LoLoL
                        :nft:meta:image = https://vertex.link/favicon.ico
                ]''')
            self.len(1, nodes)
            node = nodes[0]
            self.eq(('2bdea834252a220b61aadf592cc0de66', '30'), node.ndef[1])
            self.eq('2bdea834252a220b61aadf592cc0de66', node.get('contract'))
            self.eq('30', node.get('tokenid'))
            self.eq(('eth', 'aaaa'), node.get('owner'))
            self.eq('https://coin.vertex.link/nfts/30', node.get('nft:url'))
            self.eq({'name': 'WootWoot'}, node.get('nft:meta'))
            self.eq('WootWoot', node.get('nft:meta:name'))
            self.eq('LoLoL', node.get('nft:meta:description'))
            self.eq('https://vertex.link/favicon.ico', node.get('nft:meta:image'))

            nodes = await core.nodes('''
                [ crypto:currency:transaction=*
                    :value = '1e-24'
                ]''')
            self.len(1, nodes)
            self.eq(nodes[0].get('value'), '0.000000000000000000000001')

            nodes = await core.nodes('''
                [ crypto:currency:transaction=*
                    :value = 0.000000000000000000000002
                ]''')
            self.len(1, await core.nodes('crypto:currency:transaction:value=1e-24'))
            self.len(1, await core.nodes('crypto:currency:transaction:value=0.000000000000000000000001'))

            huge = '730750818665451459101841.00000000000000000002'
            huge2 = '730750818665451459101841.000000000000000000015'

            self.len(1, await core.nodes(f'[ crypto:currency:transaction=* :value={huge} ]'))
            self.len(1, await core.nodes(f'[ crypto:currency:transaction=* :value={huge2} ]'))
            self.len(2, await core.nodes(f'crypto:currency:transaction:value={huge}'))

    async def test_norm_lm_ntlm(self):
        async with self.getTestCore() as core:  # type: s_cortex.Cortex
            lm = core.model.type('hash:lm')
            valu, subs = lm.norm(TEST_MD5.upper())
            self.eq(valu, TEST_MD5)
            self.eq(subs, {})
            self.raises(s_exc.BadTypeValu, lm.norm, TEST_SHA256)

            ntlm = core.model.type('hash:ntlm')
            valu, subs = lm.norm(TEST_MD5.upper())
            self.eq(valu, TEST_MD5)
            self.eq(subs, {})
            self.raises(s_exc.BadTypeValu, ntlm.norm, TEST_SHA256)

    async def test_forms_crypto_simple(self):
        async with self.getTestCore() as core:  # type: s_cortex.Cortex
            async with await core.snap() as snap:
                # md5
                node = await snap.addNode('hash:md5', TEST_MD5.upper())
                self.eq(node.ndef, ('hash:md5', TEST_MD5))
                await self.asyncraises(s_exc.BadTypeValu, snap.addNode('hash:md5', TEST_SHA1))
                # sha1
                node = await snap.addNode('hash:sha1', TEST_SHA1.upper())
                self.eq(node.ndef, ('hash:sha1', TEST_SHA1))
                await self.asyncraises(s_exc.BadTypeValu, snap.addNode('hash:sha1', TEST_SHA256))
                # sha256
                node = await snap.addNode('hash:sha256', TEST_SHA256.upper())
                self.eq(node.ndef, ('hash:sha256', TEST_SHA256))
                await self.asyncraises(s_exc.BadTypeValu, snap.addNode('hash:sha256', TEST_SHA384))
                # sha384
                node = await snap.addNode('hash:sha384', TEST_SHA384.upper())
                self.eq(node.ndef, ('hash:sha384', TEST_SHA384))
                await self.asyncraises(s_exc.BadTypeValu, snap.addNode('hash:sha384', TEST_SHA512))
                # sha512
                node = await snap.addNode('hash:sha512', TEST_SHA512.upper())
                self.eq(node.ndef, ('hash:sha512', TEST_SHA512))
                await self.asyncraises(s_exc.BadTypeValu, snap.addNode('hash:sha512', TEST_MD5))

    async def test_form_rsakey(self):
        prop = 'rsa:key'
        props = {
            'bits': BITS,
            'priv:exp': HEXSTR_PRIVATE_EXPONENT,
            'priv:p': HEXSTR_PRIVATE_PRIME_P,
            'priv:q': HEXSTR_PRIVATE_PRIME_Q,
        }
        valu = (HEXSTR_MODULUS, HEXSTR_PUBLIC_EXPONENT)

        async with self.getTestCore() as core:  # type: s_cortex.Cortex

            async with await core.snap() as snap:

                node = await snap.addNode(prop, valu, props)

                self.eq(node.ndef[1], (HEXSTR_MODULUS, HEXSTR_PUBLIC_EXPONENT))

                self.eq(node.get('mod'), HEXSTR_MODULUS)
                self.eq(node.get('bits'), BITS)
                self.eq(node.get('pub:exp'), HEXSTR_PUBLIC_EXPONENT)
                self.eq(node.get('priv:exp'), HEXSTR_PRIVATE_EXPONENT)
                self.eq(node.get('priv:p'), HEXSTR_PRIVATE_PRIME_P)
                self.eq(node.get('priv:q'), HEXSTR_PRIVATE_PRIME_Q)

    async def test_model_x509(self):

        async with self.getTestCore() as core:

            crl = s_common.guid()
            cert = s_common.guid()
            icert = s_common.guid()
            fileguid = 'guid:' + s_common.guid()

            nodes = await core.nodes('''
                [ crypto:x509:cert=$icert
                    :subject="CN=issuer.link"
                    :issuer:cert=$icert
                    :selfsigned=$lib.true
                ]
            ''', opts={'vars': {'icert': icert}})
            self.eq(nodes[0].ndef, ('crypto:x509:cert', icert))
            self.eq(nodes[0].get('subject'), "CN=issuer.link")
            self.eq(nodes[0].get('issuer:cert'), icert)
            self.eq(nodes[0].get('selfsigned'), True)

            nodes = await core.nodes('''
                [ crypto:x509:cert=$cert

                    :subject="CN=vertex.link"
                    :issuer="DN FOO THING"
                    :issuer:cert=$icert

                    :serial=12345
                    :version=v3

                    :validity:notafter=2019
                    :validity:notbefore=2015

                    :md5=$md5
                    :sha1=$sha1
                    :sha256=$sha256

                    :algo=1.2.840.113549.1.1.11
                    :rsa:key=(ff00ff00, 100)
                    :signature=ff00ff00

                    :ext:sans=((dns, vertex.link), (dns, "*.vertex.link"))
                    :ext:crls = ((dns, http://vertex.link/crls),)
                    :crl:urls = ("http://vertex.link/crls",)

                    :identities:urls=(http://woot.com/1, http://woot.com/2)
                    :identities:fqdns=(vertex.link, woot.com)
                    :identities:ipv4s=(1.2.3.4, 5.5.5.5)
                    :identities:ipv6s=(ff::11, ff::aa)
                    :identities:emails=(visi@vertex.link, v@vtx.lk)
                ]
            ''', opts={'vars': {'icert': icert, 'cert': cert, 'md5': TEST_MD5, 'sha1': TEST_SHA1, 'sha256': TEST_SHA256}})

            self.eq(nodes[0].ndef, ('crypto:x509:cert', cert))
            self.eq(nodes[0].get('subject'), "CN=vertex.link")
            self.eq(nodes[0].get('issuer'), "DN FOO THING")
            self.eq(nodes[0].get('issuer:cert'), icert)
            self.eq(nodes[0].get('serial'), "12345")
            self.eq(nodes[0].get('version'), 2)

            self.eq(nodes[0].get('validity:notafter'), 1546300800000)
            self.eq(nodes[0].get('validity:notbefore'), 1420070400000)

            self.eq(nodes[0].get('md5'), TEST_MD5)
            self.eq(nodes[0].get('sha1'), TEST_SHA1)
            self.eq(nodes[0].get('sha256'), TEST_SHA256)

            self.eq(nodes[0].get('algo'), '1.2.840.113549.1.1.11')
            self.eq(nodes[0].get('rsa:key'), ('ff00ff00', 100))
            self.eq(nodes[0].get('signature'), 'ff00ff00')
            self.eq(nodes[0].get('ext:crls'), (('dns', 'http://vertex.link/crls'),))
            self.eq(nodes[0].get('crl:urls'), ('http://vertex.link/crls',))
            self.eq(nodes[0].get('ext:sans'), (('dns', '*.vertex.link'), ('dns', 'vertex.link')))
            self.eq(nodes[0].get('identities:urls'), ('http://woot.com/1', 'http://woot.com/2'))
            self.eq(nodes[0].get('identities:fqdns'), ('vertex.link', 'woot.com'))
            self.eq(nodes[0].get('identities:ipv4s'), (0x01020304, 0x05050505))
            self.eq(nodes[0].get('identities:ipv6s'), ('ff::11', 'ff::aa'))

            nodes = await core.nodes('''
                [
                    crypto:x509:crl=$crl
                        :url=http://vertex.link/crls
                        :file="*"
                ]
            ''', opts={'vars': {'crl': crl}})

            self.eq(nodes[0].ndef, ('crypto:x509:crl', crl))
            self.nn(nodes[0].get('file'))
            self.eq(nodes[0].get('url'), 'http://vertex.link/crls')

            opts = {'vars': {'cert': cert, 'file': fileguid}}
            nodes = await core.nodes('[ crypto:x509:signedfile = ($cert, $file) ]', opts=opts)

            self.eq(nodes[0].ndef, ('crypto:x509:signedfile', (cert, fileguid)))
            self.eq(nodes[0].get('cert'), cert)
            self.nn(nodes[0].get('file'), fileguid)

            opts = {'vars': {'cert': cert, 'crl': crl}}
            nodes = await core.nodes('[ crypto:x509:revoked = ($crl, $cert) ]', opts=opts)

            self.eq(nodes[0].ndef, ('crypto:x509:revoked', (crl, cert)))
            self.eq(nodes[0].get('crl'), crl)
            self.nn(nodes[0].get('cert'), cert)



import asyncio

import synapse.lib.cell as s_cell
import synapse.lib.coro as s_coro
import synapse.lib.stormsvc as s_stormsvc

import synapse.exc as s_exc
import synapse.common as s_common
import synapse.daemon as s_daemon
import synapse.telepath as s_telepath

import synapse.tests.utils as s_t_utils

class Foo:
    def woot(self):
        return 10

class DaemonTest(s_t_utils.SynTest):

    async def test_unixsock_longpath(self):

        # Explicit failure for starting a daemon with a path too deep
        # this also covers a cell failure case since the cell may start
        # a daemon.
        # This fails because of limitations onf the path length for a UNIX
        # socket being no greater than what may be stored in a mbuf.
        # The maximum length is OS dependent; with Linux using 108 characters
        # and BSD's using 104.
        with self.getTestDir() as dirn:
            extrapath = 108 * 'A'
            longdirn = s_common.genpath(dirn, extrapath)
            listpath = f'unix://{s_common.genpath(longdirn, "sock")}'
            with self.getAsyncLoggerStream('synapse.daemon',
                                           'exceeds OS supported UNIX socket path length') as stream:

                async with await s_daemon.Daemon.anit() as dmon:
                    with self.raises(OSError):
                        await dmon.listen(listpath)

                self.true(await stream.wait(1))

    async def test_dmon_ready(self):

        async with await s_daemon.Daemon.anit() as dmon:

            host, port = await dmon.listen('tcp://127.0.0.1:0')
            dmon.share('foo', Foo())

            async with await s_telepath.openurl(f'tcp://127.0.0.1:{port}/foo') as foo:
                self.eq(10, await foo.woot())
                await dmon.setReady(False)
                await foo.waitfini(timeout=2)
                self.true(foo.isfini)

            with self.raises(s_exc.LinkShutDown):
                async with await s_telepath.openurl(f'tcp://127.0.0.1:{port}/foo') as foo:
                    pass

class SvcApi(s_cell.CellApi, s_stormsvc.StormSvc):
    _storm_svc_name = 'foo'
    _storm_svc_pkgs = (  # type:  ignore
        {
            'name': 'foo',
            'version': (0, 0, 1),
            'modules': (
                {
                    'name': 'foo.mod',
                    'storm': '''
                        $x = (3)

                        function run_all() {
                            for $item in $lib.service.get($modconf.svciden).run() {
                                {}
                            }
                            return ($lib.null)
                        }

                        function run_break() {
                            for $i in $lib.service.get($modconf.svciden).run() {
                                if ($i > $x) { return($lib.null) }
                            }
                            return($lib.null)
                        }

                        function run_err() {
                            for $i in $lib.service.get($modconf.svciden).run() {
                                if ($i > $x) { [inet:newp=3] }
                            }
                            return($lib.null)
                        }
                    '''
                },
            ),
        },
    )

    async def run(self):
        async for item in self.cell.run():
            yield item


class Svc(s_cell.Cell):
    cellapi = SvcApi

    async def initServiceStorage(self):
        self.events = []

    async def run(self):
        event = asyncio.Event()
        self.events.append(event)
        try:
            for i in range(100):
                yield i
                await asyncio.sleep(0)
        finally:
            event.set()


class GenrCloseTest(s_t_utils.SynTest):

    async def test_close(self):

        async with self.getTestCoreProxSvc(Svc) as (core, core_prox, svc):

            # storm exits early
            await core.stormlist('$lib.import(foo.mod).run_break()')
            self.true(await s_coro.event_wait(svc.events[0], timeout=1))

            # storm raises part way through iterating
            await core.stormlist('$lib.import(foo.mod).run_err()')
            self.true(await s_coro.event_wait(svc.events[1], timeout=1))

            # storm normal case
            await core.stormlist('$lib.import(foo.mod).run_all()')
            self.true(await s_coro.event_wait(svc.events[2], timeout=1))

            async with svc.getLocalProxy() as svc_prox:

                # telepath exits early
                async for i in svc_prox.run():
                    if i > 3:
                        break
                self.true(await s_coro.event_wait(svc.events[3], timeout=1))

                # telepath normal case
                async for i in svc_prox.run():
                    pass
                self.true(await s_coro.event_wait(svc.events[4], timeout=1))

            # python
            async for i in svc.run():
                if i > 3:
                    break
            self.true(await s_coro.event_wait(svc.events[5], timeout=1))



import hashlib

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.msgpack as s_msgpack
import synapse.lib.provenance as s_provenance

import synapse.tests.utils as s_t_utils
from synapse.tests.utils import alist

class ProvenanceTest(s_t_utils.SynTest):

    async def test_prov(self):

        s_provenance.reset()

        async with self.getTestCore() as real, real.getLocalProxy() as core:

            # Non-existent iden
            self.none(await core.getProvStack('abcd'))

            await real.view.addTrigger({
                'cond': 'node:add',
                'form': 'test:str',
                'storm': '[ test:int=1 ]',
            })
            await s_common.aspin(core.eval('[ test:str=foo ]'))
            await self.agenlen(1, core.eval('test:int'))

            await self.agenlen(0, core.eval('test:int | delnode'))

            splices = await alist(core.splices(None, 1000))

            self.len(9, splices)
            idens = [splice[1][1].get('prov') for splice in splices]

            # source splices
            self.eq(idens[0], idens[1])
            self.eq(idens[0], idens[2])

            # test:str add
            self.eq(idens[3], idens[4])

            # trigger
            self.eq(idens[5], idens[6])

            # test:int delnode
            self.eq(idens[7], idens[8])

            provs = [await core.getProvStack(iden) for iden in idens]

            # The meta:source splices
            self.eq(({}, (('init', {'meth': '_initCoreMods'}),)), provs[0])

            # The test:str splices
            prov = provs[3][1]
            rootiden = prov[0][1]['user']
            s2 = ('storm', {'q': '[ test:str=foo ]', 'user': rootiden})
            self.eq((s2, ), prov)

            # Validate that the iden calc itself is correct
            rawprov = ({}, [('storm', (('q', '[ test:str=foo ]'), ('user', rootiden)))])
            hash = hashlib.md5(s_msgpack.en(rawprov)).hexdigest()
            self.eq(hash, idens[3])

            # The trigger splices
            prov = provs[5][1]
            s3 = ('trig', {'cond': 'node:add', 'form': 'test:str', 'tag': None, 'prop': None})
            s4 = ('storm', {'q': '[ test:int=1 ]', 'user': rootiden})
            self.eq((s2, s3, s4), prov)

            # prop:del/node:del
            prov = provs[7][1]
            ds2 = ('storm', {'q': 'test:int | delnode', 'user': rootiden})
            ds3 = ('stormcmd', {'name': 'delnode'})
            self.eq((ds2, ds3), prov)

            # Test the streaming API
            provstacks = await alist(core.provStacks(0, 1000))
            correct = [(idens[0], provs[0]), (idens[5], provs[5]), (idens[3], provs[3]), (idens[7], provs[7])]
            self.eq(provstacks, correct)

            # Force recursion exception to be thrown
            q = '.created ' + '| uniq' * 257
            with self.raises(s_exc.RecursionLimitHit) as cm:
                await real.nodes(q)

            self.eq(cm.exception.get('type'), 'stormcmd')
            self.eq(cm.exception.get('info'), {'name': 'uniq'})
            baseframe = cm.exception.get('baseframe')
            name, args = baseframe
            self.eq(name, 'storm')
            self.eq(args[0], ('q', q))
            recent_frames = cm.exception.get('recent_frames')
            self.len(6, recent_frames)
            for frame in recent_frames:
                self.eq(frame, ('stormcmd', (('name', 'uniq'),)))

            # Run a feed function and validate the user is recorded.
            await core.addFeedData('syn.nodes', [(('test:int', 1138), {})])
            # We have to brute force the last prov stack to get the data
            # Since we don't have splices to track
            stacks = await alist(core.provStacks(0, 1000))
            feed_stack = stacks[-1]
            frame = feed_stack[1][1][0]
            self.eq(frame[0], 'feed:data')
            self.eq(frame[1].get('name'), 'syn.nodes')
            self.isin('user', frame[1])

    async def test_prov_no_extra(self):
        '''
        No more than 1 prov:new event with the same data shall be fired for the same query
        '''
        self.skip('Pending provenance cache')
        async with self.getTestCore() as core:
            mesgs = await core.stormlist('[test:str=foo :hehe=bar]', opts={'editformat': 'nodeedits'})
            provs = [m for m in mesgs if m[0] == 'prov:new']

            # No duplicate prov:new
            self.len(1, provs)

    async def test_prov_disabled(self):
        '''
        Test that things still work with provenance disabled
        '''
        async with self.getTestCoreAndProxy(conf={}) as (core, prox):
            await s_common.aspin(prox.eval('[ test:str=foo ]'))

            await alist(prox.provStacks(0, 1000))

            self.none(await prox.getProvStack('abcd'))

            self.none(core.provstor.precommit())

            retn = core.provstor.stor()
            self.eq((None, None), retn)



import synapse.exc as s_exc

import synapse.lib.stormtypes as s_stormtypes

import yaml
import yaml.error

@s_stormtypes.registry.registerLib
class LibYaml(s_stormtypes.Lib):
    '''
    A Storm Library for saving/loading YAML data.
    '''
    _storm_locals = (
        {'name': 'save', 'desc': 'Encode data as a YAML string.',
         'type': {'type': 'function', '_funcname': 'save',
                  'args': (
                      {'name': 'valu', 'type': 'object', 'desc': 'The object to encode.'},
                      {'name': 'sort_keys', 'type': 'boolean', 'desc': 'Sort object keys.', 'default': True},
                  ),
                  'returns': {'type': 'str', 'desc': 'A YAML string.'}}},
        {'name': 'load', 'desc': 'Decode a YAML string/bytes into an object.',
         'type': {'type': 'function', '_funcname': 'load',
                  'args': (
                      {'name': 'valu', 'type': 'str', 'desc': 'The string to decode.'},
                  ),
                  'returns': {'type': 'prim', 'desc': 'The decoded primitive object.'}}},
    )
    _storm_lib_path = ('yaml',)

    def getObjLocals(self):
        return {
            'save': self.save,
            'load': self.load,
        }

    async def save(self, valu, sort_keys=True):
        valu = await s_stormtypes.toprim(valu)
        sort_keys = await s_stormtypes.tobool(sort_keys)
        return yaml.safe_dump(valu, sort_keys=sort_keys)

    async def load(self, valu):
        valu = await s_stormtypes.tostr(valu)
        try:
            return yaml.safe_load(valu)
        except yaml.error.YAMLError as e:
            raise s_exc.BadArg(mesg=f'Invalid YAML text: {str(e)}')



import synapse.exc as s_exc
import synapse.common as s_common

import synapse.tests.utils as s_t_utils

class BaseTest(s_t_utils.SynTest):

    async def test_model_base_timeline(self):

        async with self.getTestCore() as core:

            nodes = await core.nodes('[ meta:timeline=* :title=Woot :summary=4LOLZ :type=lol.cats ]')
            self.len(1, nodes)
            nodes = await core.nodes('''
                [ meta:event=* :title=Zip :duration=1:30:00
                    :summary=Zop :time=20220321 :type=zip.zop :timeline={meta:timeline:title=Woot} ]''')
            self.len(1, nodes)
            nodes = await core.nodes('''[ meta:event=* :title=Hehe :duration=2:00
                    :summary=Haha :time=20220322 :type=hehe.haha :timeline={meta:timeline:title=Woot} ]''')
            self.len(1, nodes)

            self.len(2, await core.nodes('meta:timeline +:title=Woot +:summary=4LOLZ +:type=lol.cats -> meta:event'))
            self.len(1, await core.nodes('meta:timeline -> meta:timeline:taxonomy'))
            self.len(2, await core.nodes('meta:event -> meta:event:taxonomy'))
            self.len(1, await core.nodes('meta:event +:title=Hehe +:summary=Haha +:time=20220322 +:duration=120 +:type=hehe.haha +:timeline'))

    async def test_model_base_note(self):

        async with self.getTestCore() as core:
            nodes = await core.nodes('[ inet:fqdn=vertex.link inet:fqdn=woot.com ] | note.add "foo bar baz"')
            self.len(2, nodes)
            self.len(1, await core.nodes('meta:note'))
            self.len(1, await core.nodes('meta:note:created<=now'))
            self.len(1, await core.nodes('meta:note:creator=$lib.user.iden'))
            self.len(1, await core.nodes('meta:note:text="foo bar baz"'))
            self.len(2, await core.nodes('meta:note -(about)> inet:fqdn'))
            self.len(1, await core.nodes('meta:note [ :author={[ ps:contact=* :name=visi ]} ]'))
            self.len(1, await core.nodes('ps:contact:name=visi -> meta:note'))

            # Notes are always unique when made by note.add
            nodes = await core.nodes('[ inet:fqdn=vertex.link inet:fqdn=woot.com ] | note.add "foo bar baz"')
            self.len(2, await core.nodes('meta:note'))
            self.ne(nodes[0].ndef, nodes[1].ndef)
            self.eq(nodes[0].get('text'), nodes[1].get('text'))

    async def test_model_base_node(self):

        async with self.getTestCore() as core:

            async with await core.snap() as snap:

                iden = s_common.guid()

                props = {
                    'type': 'hehe haha',
                    'data': ('some', 'data', 'here'),
                }

                node = await snap.addNode('graph:node', iden, props=props)
                self.eq(node.ndef, ('graph:node', iden))
                self.eq(node.get('type'), 'hehe haha')
                self.eq(node.get('data'), ('some', 'data', 'here'))

    async def test_model_base_link(self):

        async with self.getTestCore() as core:

            async with await core.snap() as snap:

                node1 = await snap.addNode('test:int', 20)
                node2 = await snap.addNode('test:str', 'foo')

                link = await snap.addNode('graph:edge', (node1.ndef, node2.ndef))

                self.eq(link.ndef[1], (('test:int', 20), ('test:str', 'foo')))
                self.eq(link.get('n1'), ('test:int', 20))
                self.eq(link.get('n1:form'), 'test:int')

                self.eq(link.get('n2'), ('test:str', 'foo'))
                self.eq(link.get('n2:form'), 'test:str')

                timeedge = await snap.addNode('graph:timeedge', (node1.ndef, node2.ndef, '2015'))

                self.eq(timeedge.ndef[1], (('test:int', 20), ('test:str', 'foo'), 1420070400000))

                self.eq(timeedge.get('time'), 1420070400000)

                self.eq(timeedge.get('n1'), ('test:int', 20))
                self.eq(timeedge.get('n1:form'), 'test:int')

                self.eq(timeedge.get('n2'), ('test:str', 'foo'))
                self.eq(timeedge.get('n2:form'), 'test:str')

    async def test_model_base_event(self):

        async with self.getTestCore() as core:

            async with await core.snap() as snap:

                iden = s_common.guid()

                props = {
                    'type': 'HeHe HaHa',
                    'time': '2015',
                    'name': 'Magic Pony',
                    'data': ('some', 'data', 'here'),
                }

                node = await snap.addNode('graph:event', iden, props=props)

                self.eq(node.ndef, ('graph:event', iden))

                self.eq(node.get('type'), 'HeHe HaHa')
                self.eq(node.get('time'), 1420070400000)
                self.eq(node.get('data'), ('some', 'data', 'here'))
                self.eq(node.get('name'), 'Magic Pony')

                # Raise on non-json-safe values
                props['data'] = {(1, 2): 'foo'}
                await self.asyncraises(s_exc.BadTypeValu, snap.addNode('graph:event', iden, props=props))

                props['data'] = b'bindata'
                await self.asyncraises(s_exc.BadTypeValu, snap.addNode('graph:event', iden, props=props))

    async def test_model_base_edge(self):

        async with self.getTestCore() as core:

            pers = s_common.guid()
            plac = s_common.guid()

            n1def = ('ps:person', pers)
            n2def = ('geo:place', plac)

            async with await core.snap() as snap:

                node = await snap.addNode('edge:has', (n1def, n2def))

                self.eq(node.get('n1'), n1def)
                self.eq(node.get('n1:form'), 'ps:person')

                self.eq(node.get('n2'), n2def)
                self.eq(node.get('n2:form'), 'geo:place')

                node = await snap.addNode('edge:wentto', (n1def, n2def, '2016'))

                self.eq(node.get('time'), 1451606400000)

                self.eq(node.get('n1'), n1def)
                self.eq(node.get('n1:form'), 'ps:person')

                self.eq(node.get('n2'), n2def)
                self.eq(node.get('n2:form'), 'geo:place')

            opts = {'vars': {'pers': pers}}

            await self.agenlen(1, core.eval('ps:person=$pers -> edge:has -> *', opts=opts))
            await self.agenlen(1, core.eval('ps:person=$pers -> edge:has -> geo:place', opts=opts))
            await self.agenlen(0, core.eval('ps:person=$pers -> edge:has -> inet:ipv4', opts=opts))

            await self.agenlen(1, core.eval('ps:person=$pers -> edge:wentto -> *', opts=opts))
            q = 'ps:person=$pers -> edge:wentto +:time@=(2014,2017) -> geo:place'
            await self.agenlen(1, core.eval(q, opts=opts))
            await self.agenlen(0, core.eval('ps:person=$pers -> edge:wentto -> inet:ipv4', opts=opts))

            opts = {'vars': {'place': plac}}

            await self.agenlen(1, core.eval('geo:place=$place <- edge:has <- *', opts=opts))
            await self.agenlen(1, core.eval('geo:place=$place <- edge:has <- ps:person', opts=opts))
            await self.agenlen(0, core.eval('geo:place=$place <- edge:has <- inet:ipv4', opts=opts))

            # Make a restricted edge and validate that you can only form certain relationships
            copts = {'n1:forms': ('ps:person',), 'n2:forms': ('geo:place',)}
            t = core.model.type('edge').clone(copts)
            norm, info = t.norm((n1def, n2def))
            self.eq(norm, (n1def, n2def))
            self.raises(s_exc.BadTypeValu, t.norm, (n1def, ('test:int', 1)))
            self.raises(s_exc.BadTypeValu, t.norm, (('test:int', 1), n2def))

            # Make sure we don't return None nodes if one node of an edge is deleted
            node = await core.getNodeByNdef(n2def)
            await node.delete()
            opts = {'vars': {'pers': pers}}
            await self.agenlen(0, core.eval('ps:person=$pers -> edge:wentto -> *', opts=opts))

            # Make sure we don't return None nodes on a PropPivotOut
            opts = {'vars': {'pers': pers}}
            await self.agenlen(0, core.eval('ps:person=$pers -> edge:wentto :n2 -> *', opts=opts))

    async def test_model_base_source(self):

        async with self.getTestCore() as core:

            async with await core.snap() as snap:

                props = {
                    'name': 'FOO BAR',
                    'type': 'osint',
                }

                sorc = await snap.addNode('meta:source', '*', props=props)

                self.eq(sorc.get('type'), 'osint')
                self.eq(sorc.get('name'), 'foo bar')

                valu = (sorc.ndef[1], ('inet:fqdn', 'woot.com'))

                seen = await snap.addNode('meta:seen', valu)

                self.eq(seen.get('source'), sorc.ndef[1])
                self.eq(seen.get('node'), ('inet:fqdn', 'woot.com'))

    async def test_model_base_cluster(self):

        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                guid = s_common.guid()
                props = {'name': 'Test Cluster', 'desc': 'A cluster for testing', 'type': 'similarity'}
                cnode = await snap.addNode('graph:cluster', guid, props)
                self.eq(cnode.get('type'), 'similarity')
                self.eq(cnode.get('name'), 'test cluster')
                self.eq(cnode.get('desc'), 'a cluster for testing')

                # Example reference nodes
                r1 = await snap.addNode('edge:refs', (cnode.ndef, ('test:str', '1234')))
                r2 = await snap.addNode('edge:refs', (cnode.ndef, ('test:int', 1234)))

                # Gather up all the nodes in the cluster
                nodes = await snap.eval(f'graph:cluster={guid} -+> edge:refs -+> * | uniq').list()
                self.len(5, nodes)



import synapse.lib.cell as s_cell
import synapse.lib.stormsvc as s_stormsvc

class TestsvcApi(s_cell.CellApi, s_stormsvc.StormSvc):
    _storm_svc_name = 'testsvc'
    _storm_svc_vers = (0, 0, 1)
    _storm_svc_pkgs = (
        {
            'name': 'testsvc',
            'version': (0, 0, 1),
            'commands': (
                {
                    'name': 'testsvc.test',
                    'storm': '$lib.print($lib.service.get($cmdconf.svciden).test())',
                },
            )
        },
    )

    async def test(self):
        return await self.cell.test()

class Testsvc(s_cell.Cell):

    cellapi = TestsvcApi

    confdefs = {
        'secret': {'type': 'string'},
    }

    async def __anit__(self, dirn, conf):
        await s_cell.Cell.__anit__(self, dirn, conf=conf)
        self.secret = self.conf.reqConfValu('secret')

    async def test(self):
        return self.secret



import math
import synapse.tests.utils as s_t_utils

import synapse.lib.gis as s_gis

# earth mean radius in mm
r = 6371008800

ratios = {
    'cm': 10.0,
    'm': 1000.0,
    'km': 1000000.0,
}

km = 1000000.0  # using mm as base units
gchq = (51.8994, -2.0783)

class GisTest(s_t_utils.SynTest):

    def test_lib_gis_haversine(self):
        px = (36.12, -86.67)
        py = (33.94, -118.40)
        self.eqish(s_gis.haversine(px, py), 2886448429.7648544)

        # Test haversinve value from rosetta code
        r = s_gis.haversine((36.12, -86.67), (33.94, -118.40), 6372.8)
        e = 2887.2599506071106
        self.eqish(r, e)

        # Test against 1/4th of a unit sphere
        r = s_gis.haversine((45, 45), (-45, 45), 1.0)
        e = math.pi / 2
        # We are typically within the machine-epsilon range for this test
        self.assertAlmostEqual(r, e)

        # Test against the haversine package
        lyon = (45.7597, 4.8422)
        paris = (48.8567, 2.3508)
        r = s_gis.haversine(lyon, paris, r=6371)
        e = 392.21671780659625
        self.assertAlmostEqual(r, e)

    def test_lib_gis_latlong(self):
        self.assertRaises(ValueError, s_gis.latlong, 'hehe')
        self.assertRaises(ValueError, s_gis.latlong, 'hehe,hoho')

        self.eq(s_gis.latlong(' 0,0 '), (0.0, 0.0))
        self.eq(s_gis.latlong('-0,0 '), (0.0, 0.0))
        self.eq(s_gis.latlong('50,100 '), (50.0, 100.0))
        self.eq(s_gis.latlong('-50,100 '), (-50.0, 100.0))
        self.eq(s_gis.latlong('-50,-100 '), (-50.0, -100.0))
        self.eq(s_gis.latlong('50,-100 '), (50.0, -100.0))

        self.eq(s_gis.latlong(' 0.12345678901234567890,-0.12345678901234567890 '), (0.12345678901234568, -0.12345678901234568))  # Note precision

        self.eq(s_gis.latlong('123.456,-987.654 '), (123.456, -987.654))  # Note Invalid coords

    def test_lib_gis_near(self):
        point = (0.0, 0.0)
        dist = 0
        points = []
        self.false(s_gis.near(point, dist, points))  # no points

        point = (0.0, 0.0)
        dist = 0
        points = [(0.0, 0.0)]
        self.true(s_gis.near(point, dist, points))  # same point

        point = (0.0, 0.0)
        dist = 0
        points = [(50.0, 50.0), (0.0, 0.0)]
        self.true(s_gis.near(point, dist, points))  # far point and same point

        point = (45.7597, 4.8422)  # lyon
        dist = 400000000  # actual haversine distance between lyon/paris is ~392217259mm
        points = [(0.0, 0.0), (48.8567, 2.3508)]  # 0,0 and paris
        self.true(s_gis.near(point, dist, points))

        point = (45.7597, 4.8422)  # lyon
        dist = 391000000  # actual haversine distance between lyon/paris is ~392217259mm
        points = [(0.0, 0.0), (48.8567, 2.3508)]  # 0,0 and paris
        self.false(s_gis.near(point, dist, points))

    def test_lib_gis_dms2dec(self):
        self.eqish(s_gis.dms2dec(45, 46, 52), 45.78111111111111)

    def test_lib_gis_bbox(self):
        lbox = s_gis.bbox(gchq[0], gchq[1], 1 * km)
        self.eq(lbox, (51.890406796362754,
                       51.908393203637246,
                       -2.0928746526154747,
                       -2.0637253473845254))
